import os
import simfile
import torch
from torch.utils.data import Dataset
import pickle
from torchvision.io import read_image
from preliminary_experiment import get_middle_chart


class SongPacksDataset(Dataset):
    #def __init__(self, annotations_file, img_dir, transform=None, target_transform=None):
    def __init__(self, img_paths, img_waves, img_sample_rates, difficulties, charts,
                 mel_specs_23, mel_specs_46, mel_specs_93, device, indexed_vocabulary, transform=None, target_transform=None):
        # for each song, 3 tuple of images as tensors
        #self.all_paths = img_paths
        #self.all_waves = img_waves
        #self.all_sample_rates = img_sample_rates
        #self.all_difficulties = difficulties
        #self.all_charts = charts
        #self.mel_specs_23 = mel_specs_23
        #self.mel_specs_46 = mel_specs_46
        #self.mel_specs_93 = mel_specs_93
        #self.device = device
        #self.transform = transform
        #self.target_transform = target_transform
        # imagenet loading
        # all same numbers, all equally sorted and chosen
        self.all_paths_folder = img_paths
        self.all_paths = os.listdir(img_paths)
        # self.all_paths = sorted(self.all_paths, key=custom_sort_key)
        self.all_waves_folder = img_waves
        self.all_waves = os.listdir(img_waves)
        self.all_sample_rates_folder = img_sample_rates
        self.all_sample_rates = os.listdir(img_sample_rates)
        self.all_difficulties_folder = difficulties
        self.all_difficulties = os.listdir(difficulties)
        self.all_charts = os.listdir(charts)
        self.all_charts_folder = charts
        self.mel_specs_23_folder = mel_specs_23
        self.mel_specs_23 = os.listdir(mel_specs_23)
        self.mel_specs_46_folder = mel_specs_46
        self.mel_specs_46 = os.listdir(mel_specs_46)
        self.mel_specs_93_folder = mel_specs_93
        self.mel_specs_93 = os.listdir(mel_specs_93)
        self.device = device

        self.start_token = indexed_vocabulary.get("<s>")
        self.end_token = indexed_vocabulary.get("<e>")

        assert (len(self.all_paths) == len(self.all_charts) == len(self.all_waves) == len(self.all_sample_rates)
                == len(self.all_difficulties) == len(self.mel_specs_23)
                == len(self.mel_specs_46) == len(self.mel_specs_93))

    def __len__(self):
        #return len(os.listdir(self.all_charts))
        return len(self.all_charts)

    # get items(a song) gives back 3 images, difficulty and chart
    def __getitem__(self, idx):
        #img_path = os.path.join(self.img_dir, self.img_labels.iloc[idx, 0])
        #image = read_image(img_path)
        #label = self.img_labels.iloc[idx, 1]
        #if self.transform:
        #    image = self.transform(image)
        #if self.target_transform:
        #    label = self.target_transform(label)
        #with open(f"{self.all_paths}{idx}.txt", 'r') as f:
        #    song_path = f.read()
        with open(f"{self.all_paths_folder}{self.all_paths[idx]}", 'r') as f:
            song_path = f.read()

        #with open(f"{self.all_waves}{idx}.pkl", 'rb') as fp:
        #    song_wave = pickle.load(fp)#.to(self.device)  # maybe problem if different place saved..
        with open(f"{self.all_waves_folder}{self.all_waves[idx]}", 'rb') as fp:
            song_wave = pickle.load(fp)  # .to(self.device)  # maybe problem if different place saved..

        #song_sample_rate = torch.load(f"{self.all_sample_rates}{idx}.pt")  # .cuda und model
        #song_difficulty = torch.load(f"{self.all_difficulties}{idx}.pt")
        # song_sample_rate = torch.load(f"{self.all_sample_rates}{idx}.pt")  # .cuda und model
        song_sample_rate = torch.load(
            f"{self.all_sample_rates_folder}{self.all_sample_rates[idx]}")  # .cuda und model
        # song_difficulty = torch.load(f"{self.all_difficulties}{idx}.pt")
        song_difficulty = torch.load(f"{self.all_difficulties_folder}{self.all_difficulties[idx]}")

        #with open(f"{self.all_charts}{idx}.pkl", 'rb') as fp:
            # with open(f'/scratch/grzonkow/vocabulary.pkl', 'rb') as fp:
        #    song_chart = pickle.load(fp)#.to(self.device)

        # with open(f"{self.all_charts}{idx}.pkl", 'rb') as fp:
        with open(f"{self.all_charts_folder}{self.all_charts[idx]}", 'rb') as fp:
            # with open(f'/scratch/grzonkow/vocabulary.pkl', 'rb') as fp:
            song_chart = pickle.load(fp)  # .to(self.device)
        # wrong saved, lost start end token, have None: fix fast: replace first and last token(None) with vocab entry
        song_chart[0] = self.start_token
        song_chart[-1] = self.end_token

        #song_mel_spec_23 = torch.load(f"{self.mel_specs_23}{idx}.pt")
        #song_mel_spec_46 = torch.load(f"{self.mel_specs_46}{idx}.pt")
        #song_mel_spec_93 = torch.load(f"{self.mel_specs_93}{idx}.pt")
        # song_mel_spec_23 = torch.load(f"{self.mel_specs_23}{idx}.pt")
        song_mel_spec_23 = torch.load(f"{self.mel_specs_23_folder}{self.mel_specs_23[idx]}")
        # song_mel_spec_46 = torch.load(f"{self.mel_specs_46}{idx}.pt")
        song_mel_spec_46 = torch.load(f"{self.mel_specs_46_folder}{self.mel_specs_46[idx]}")
        # song_mel_spec_93 = torch.load(f"{self.mel_specs_93}{idx}.pt")
        song_mel_spec_93 = torch.load(f"{self.mel_specs_93_folder}{self.mel_specs_93[idx]}")

        return song_path, song_wave, song_sample_rate, song_difficulty, \
               song_chart, song_mel_spec_23, song_mel_spec_46, song_mel_spec_93