# from https://pytorch.org/audio/main/tutorials/audio_feature_extractions_tutorial.html#
# sphx-glr-tutorials-audio-feature-extractions-tutorial-py
import os
import torch
import torchaudio
import matplotlib.pyplot as plt
import librosa
import torchaudio.transforms as T


# from https://pytorch.org/audio/main/tutorials/audio_feature_extractions_tutorial.html#
# sphx-glr-tutorials-audio-feature-extractions-tutorial-py
def plot_waveform(waveform, sr, title="Waveform", ax=None):
    waveform = waveform.numpy()

    num_channels, num_frames = waveform.shape
    time_axis = torch.arange(0, num_frames) / sr

    if ax is None:
        _, ax = plt.subplots(num_channels, 1)
    ax.plot(time_axis, waveform[0], linewidth=1)
    ax.grid(True)
    ax.set_xlim([0, time_axis[-1]])
    ax.set_title(title)


# from https://pytorch.org/audio/main/tutorials/audio_feature_extractions_tutorial.html#
# sphx-glr-tutorials-audio-feature-extractions-tutorial-py
def plot_spectrogram(specgram, title=None, ylabel="freq_bin", ax=None):
    if ax is None:
        _, ax = plt.subplots(1, 1)
    if title is not None:
        ax.set_title(title)
    ax.set_ylabel(ylabel)
    ax.imshow(librosa.power_to_db(specgram), origin="lower", aspect="auto", interpolation="nearest")


def mixing_down(waveform):
    # if not mono
    if waveform.shape[0]>1:
        waveform = torch.mean(waveform, dim=0, keepdim=True)
    return waveform


def sampling_down(waveform, samplerate, targetsamplerate):
    if samplerate != targetsamplerate:
        resampler = torchaudio.transforms.Resample(samplerate, targetsamplerate)
        waveform = resampler(waveform)
    return waveform


if __name__ == '__main__':
    # test
    #song = "../../7guys1pack/ALONE [In The Dark]/ALONE [In The Dark].ogg"
    #filename = "alone_in_the_dark"

    # load pack
    path_pack = "../../test_pack/"
    songs = []
    songs_folder = []
    # target_sample_rate = ...

    # make a list of songs of file type ogg and their folder paths
    for folder in os.listdir(path_pack):
        path_folder = os.path.join(path_pack, folder)
        if os.path.isdir(path_folder):
            for file in os.listdir(path_folder):
                if file.endswith(".ogg"):
                    songs.append(f'{path_folder}/{file}')
                    songs_folder.append(path_folder)
        else:
            break
    print(songs)
    print(songs_folder)



    for i, song in enumerate(songs):
        # load audio
        waveform, sample_rate = torchaudio.load(song)
        #waveform = sampling_down(waveform, sample_rate, target_sample_rate)
        waveform = mixing_down(waveform)

        # in ms
        win_length_list_ms = [23, 46, 93]
        win_length_list = []
        for length in win_length_list_ms:
            win_length_list.append(int(sample_rate * (length / 1000)))

        # mel spectrogram transform
        for j, win_length in enumerate(win_length_list):

            # win_length = int(sample_rate * (23 / 1000))  # 23ms, 46ms, and 93m and int(sample_rate * (length / 1000))
            hop_length = int(sample_rate * (10 / 1000))  # 10ms aka stride
            n_mels = 80  # Number of mel filterbanks 80
            n_fft = win_length

            # https://pytorch.org/audio/stable/generated/torchaudio.functional.melscale_fbanks.html#torchaudio.functional.melscale_fbanks
            # https://pytorch.org/audio/stable/generated/torchaudio.transforms.MelSpectrogram.html#torchaudio.transforms.MelSpectrogram
            mel_spectrogram_transform = T.MelSpectrogram(
                sample_rate=sample_rate,
                n_fft=n_fft,
                win_length=win_length,
                hop_length=hop_length,
                center=True,
                pad_mode="reflect",
                power=2.0,
                norm="slaney",
                n_mels=n_mels,
                mel_scale="htk",
            )

            # Perform transform
            # Mel frequency spectrogram of size (…, n_mels, time) first channels?
            mel_spectrogram = mel_spectrogram_transform(waveform)
            #torch.set_printoptions(profile="full")
            #f = open(f"{songs_folder[i]}/mel_spectrogram_{win_length_list_ms[j]}.txt", "a")
            #f.write(f"{print(mel_spectrogram)}")
            #f.close()
            # pt data ending
            torch.save(mel_spectrogram, f"{songs_folder[i]}/mel_spectrogram_{win_length_list_ms[j]}.pt")
            #print(torch.load(f"{songs_folder[i]}/mel_spectrogram_{win_length_list_ms[j]}"))

            # plot_waveform(waveform, sample_rate, title="Original waveform")
            # plt.savefig(f"{songs_folder[i]}/original_waveform_{filename}.png")
            # plt.close()

            #plt.figure()
            #plot_spectrogram(mel_spectrogram[0], title="MelSpectrogram", ylabel="mel freq")
            # save spectograms in original folders
            # plt.savefig(f"./outputs/specs/mel_spectogram_{filename}.png")
            # extract pytorch tensor
            #plt.savefig(f"{songs_folder[i]}/mel_spectogram_{win_length_list_ms[j]}.png")
            #plt.clf()

