# from https://deeplearning.neuromatch.io/projects/ComputerVision/spectrogram_analysis.html
# https://pytorch.org/tutorials/beginner/transformer_tutorial.html
import argparse
from functools import partial

import torch

from preprocess import *
from sklearn.metrics import confusion_matrix


# training for each batch
def train_one_epoch(danceformer, train_loader, epoch, model_params_path, device,
                    details_dir, model_params_path_nan_problem):

    total_loss = 0.
    total_loss_v_2 = 0.
    log_interval = 1
    start_time = time.time()
    num_iters = len(train_loader)
    initial_start_time = time.time()

    empty_token = danceformer.indexed_vocabulary.get("0000")

    # Run training
    for i, data in enumerate(train_loader):
        # data consists of (data_imgs, diffs, charts) each with dim[0] = batch_size
        # for 5 seconds (in dataloader loaded):
        # data images shape: torch.Size([2, 20, 480])
        # diffs shape: torch.Size([2, 20])
        # charts shape: torch.Size([2, 320])

        # progress with batch/num_batches and elapsed time since training of epoch started
        now = datetime.now()
        iter_num = i  # % num_iters ,since per epoch, modulo not really needed
        elapsed_time = time.time() - initial_start_time
        print(f'Progress: {iter_num}/{num_iters} | Elapsed time: {elapsed_time:5.2f}s | Timestamp: {now} ')


        #load data
        data_imgs, diffs, charts = data
        data_imgs = data_imgs.to(device)
        diffs = diffs.to(device)
        charts = charts.to(device)

        # sets pre_conv_imgs, diff_reps
        data_imgs, diffs = danceformer.forward(data_imgs, diffs)
        data_imgs = data_imgs.to(device)
        diffs = diffs.to(device)
        charts = charts.to(device)
        trans_model = danceformer.trans_model.to(device)

        danceformer.optimizer.zero_grad()
        output = trans_model(data_imgs, diffs, charts).to(device)
        output_flat = output.view(-1, danceformer.ntokens).to(device)
        charts = charts[:, 1:]
        # charts_flat = charts.view(-1).to(device)
        charts_flat = charts.reshape(charts.shape[0] * charts.shape[1])

        # charts[:, 1:]
        loss = danceformer.criterion(output_flat, charts_flat)

        #loss.requires_grad_(True)
        #loss.retain_grad()
        loss.backward(retain_graph=True)
        #print(loss.grad)
        #torch.nn.utils.clip_grad_norm_(trans_model.parameters(), 0.5)
        danceformer.optimizer.step()

        # -----------------------------------------------------------------------------------
        # Metrics to check status
        #max_values = torch.max(output, dim=2)
        #predictions = torch.argmax(output, dim=1)
        #predictions = max_values.indices

        predictions = torch.argmax(output_flat, dim=1)

        # check how probability per token develops
        m = torch.nn.Softmax(dim=1).to(device)
        # 1. variante, verhältnis nicht leere zu leere tokens
        output_softmax = m(output_flat).to(device)
        # 2.variante, verhältnis richtiges token zu nicht leere tokens
        # take out empty token
        output_v_2_flat = torch.cat((output_flat[:, :empty_token], output_flat[:, empty_token + 1:]), dim=1).to(device)
        output_var_2_softmax = m(output_v_2_flat).to(device)
        # insert empty token with 0 prob as placeholer to have right index
        tmp = torch.cat((output_var_2_softmax[:, :empty_token], torch.zeros((output_var_2_softmax.shape[0]), 1, device=device)), dim=1).to(device)
        output_var_2_softmax = torch.cat((tmp[:, :empty_token+1], output_var_2_softmax[:, empty_token:]), dim=1).to(device)
        # richtiges token siehte target/chart an entsprechender stelle
        right_token_probs = output_var_2_softmax[torch.arange(output_var_2_softmax.size(0)), charts_flat].unsqueeze(1).to(device)
        # take entries where non empty chart values
        non_empty_ind = torch.where(charts_flat != empty_token, 1, 0)
        non_empty_ind_pred = torch.where(predictions != empty_token, 1, 0)
        # only right tokens prob ohne empty token
        right_token_probs = right_token_probs[non_empty_ind]
        if len(right_token_probs) != 0:
            average_right_tokens = (torch.sum(right_token_probs)/len(right_token_probs)).to(device)
        # andere token ohne leeres/eh gleich 0, also durch summe aller probs ohne right token
        rest_token_probs = ((torch.sum(output_var_2_softmax, 1)[non_empty_ind].unsqueeze(1))-right_token_probs).to(device)
        greater = torch.where(right_token_probs > rest_token_probs, 1, 0).to(device)
        # verhältnis richtige token zu falschen token
        if not (rest_token_probs == 0).all().item():
            ratio = right_token_probs/rest_token_probs

            # for greater values how much greater
            greater_ratio_values = ratio.squeeze(1)[greater.squeeze(1) == 1]
        else:
            greater_ratio_values = torch.tensor([])

        if len(non_empty_ind) != 0:
            charts_wo_empty = charts_flat[non_empty_ind == 1]
            output_wo_empty = output_flat[non_empty_ind == 1]

            # leave out empty tokens, just for evaluation, not for backpropagation
            with torch.no_grad():
                loss_v_2 = danceformer.criterion(output_wo_empty, charts_wo_empty)
        # special case, only has empty tokens, no non empty tokens -> loss_v_2 0
        else:
            loss_v_2 = 0

        # bilde summe über vocabulary pro batch*len(chart)
        # -> (batch*len(chart), 1)  zb (640, 1)
        all_probs = torch.sum(output_softmax, 1)
        # hol die probability der nicht leeren token pro (batch*len(chart))
        # -> (batch*len(chart), 1) zb (640, 1)
        empty_token_probs = output_softmax[:, empty_token]
        # für jeden zeitpunkt non empty prob -> zb(640, 1)
        all_non_empty_probs = all_probs - empty_token_probs
        # über alle zeitpunkte, summe non empty prob
        sum_all_non_empty_probs = sum(all_non_empty_probs)
        # (batch*len(chart)) anzahl zeitpunkte zb 640/639?
        number_non_empty_entries = len(output_softmax[0])-1
        avg_prob_non_empty = sum_all_non_empty_probs / number_non_empty_entries
        #avg_prob_non_empty.append(avg_prob_non_empty.item())


        right_predictions = torch.where(predictions == charts_flat, 1.0, 0.0)
        all_right_predictions = torch.tensor([1.0] * len(charts_flat))
        # usual accuracy, number right tokens from all tokens
        accuracy = sum(right_predictions) / len(predictions)

        # accuracy v2: #right predictions&non empty/#all predictions&non empty
        right_predictions_v_2 = torch.where(predictions[non_empty_ind == 1] == charts_flat[non_empty_ind == 1], 1.0, 0.0)
        all_right_predictions_v = torch.tensor([1.0] * len(charts_flat[non_empty_ind == 1]))
        # look if any prediction non empty is there, test for non empty entries of original, if they are non empty

        right_predictions_v_3 = torch.where(predictions[non_empty_ind == 1] != empty_token, 1.0, 0.0)

        # non empty tokens, right token
        if len(non_empty_ind) != 0.0:
            accuracy_v_2 = sum(right_predictions_v_2) / len(predictions[non_empty_ind == 1])
            # non empty tokens, any value besides empty (and empty constellation)
            accuracy_v_3 = sum(right_predictions_v_3) / len(predictions[non_empty_ind == 1])
        else:
            accuracy_v_2 = 0
            accuracy_v_3 = 0

        empty_check = torch.where(predictions != empty_token, 1.0, 0.0)
        empty_check_orig = torch.where(charts_flat != empty_token, 1.0, 0.0)
        #print(empty_check)
        #print(f"Number of non empty chart entries: {len(torch.nonzero(empty_check))}")




        # -----------------------------------------------------------------------------------


        # if tensor: loss.item()
        total_loss += loss
        total_loss_v_2 += loss_v_2
        if i % log_interval == 0:
            lr = danceformer.scheduler.get_last_lr()[0]
            ms_per_batch = (time.time() - start_time) * 1000 / log_interval
            # check
            cur_loss = total_loss / log_interval
            ppl = math.exp(cur_loss)
            # ppl für nicht leere token
            cur_loss_v_2 = total_loss_v_2 / log_interval
            ppl_v_2 = math.exp(cur_loss_v_2)

            # empty, non empty
            cm = confusion_matrix(non_empty_ind.cpu(), non_empty_ind_pred.cpu())
            # right token, wrong token, accuracy
            cm_rt = confusion_matrix(all_right_predictions.cpu(), right_predictions.cpu())
            # empty token position, right token
            cm_et_rt = confusion_matrix(all_right_predictions_v.cpu(), right_predictions_v_2.cpu())
            # empty token position, any non empty token
            cm_et_at = confusion_matrix(all_right_predictions_v.cpu(), right_predictions_v_3.cpu())


            print(f'| epoch {epoch:3d} | {(i+1):5d}/{num_iters:5d} batches | '
                  f'lr {lr:02.2f} | ms/batch {ms_per_batch:5.2f} | '
                  f'loss {cur_loss.item():5.2f} | ppl {ppl:8.2f} | '
                  f'accuracy {accuracy.item():5.2f} | '
                  f'loss_v_2 {cur_loss_v_2.item():5.2f} | ppl_v_2 {ppl_v_2:8.2f} | '
                  f'accuracy_v_2 (non_empty, right token) {accuracy_v_2.item():5.2f} | ' 
                  f'accuracy_v_3 (non empty, any non empty token) {accuracy_v_3.item():5.2f} | '
                  f'| # non_empty {torch.count_nonzero(empty_check)} / {torch.count_nonzero(empty_check_orig)} ')
            #print(f"Number of non empty chart entries: {len(torch.nonzero(empty_check))}")

            # save features per log_interval (for 1: batch)
            # need to be set as a new file in beginning, before epoch training
            # save features for each epoch in file
            def save_to_txt_file(feature_name, feature):
                with open(f"{details_dir}{feature_name}.txt", 'a') as fp:
                    fp.write(str(feature) + '\n')

            #save_to_txt_file("avg_prob_non_empty", avg_prob_non_empty.item())
            save_to_txt_file("ms_per_batch", ms_per_batch)
            save_to_txt_file("ppl", ppl)
            save_to_txt_file("ppl_v_2", ppl_v_2)
            save_to_txt_file("loss", cur_loss.item())
            save_to_txt_file("loss_v_2", cur_loss_v_2.item())
            save_to_txt_file("accuracy", accuracy.item())
            save_to_txt_file("accuracy_v_2", accuracy_v_2.item())
            save_to_txt_file("accuracy_v_3", accuracy_v_3.item())

            #save_to_txt_file("average_right_tokens", average_right_tokens.item())

            #print(f"Average probability non empty token per batch: {avg_prob_non_empty}")

            #if len(greater_ratio_values) != 0:
                #print(f"Greater values (ratio): {[t.item() for t in greater_ratio_values]}")
            #    greater_ratio_values_to_write = str([t.item() for t in greater_ratio_values])
            #    save_to_txt_file("greater_ratio_values", greater_ratio_values_to_write)
            #else:
            #    # gute wahl??
            #    save_to_txt_file("greater_ratio_values", str([]))

            total_loss = 0
            total_loss_v_2 = 0
            start_time = time.time()

        if i % 15 == 0:
            print(f'confusion matrix (empty, non empty) \n{cm}')
            print(f'confusion matrix rt (right token, not right token) \n{cm_rt}')
            print(f'confusion matrix et_rt, where non empty token (right token, not right token) \n{cm_et_rt}')
            print(f'confusion matrix et_at, where on empty token (non empty token, empty token) \n{cm_et_at}')

        if i % 50 == 0:
            def open_txt_file(feature_name, feature):
                with open(f"{details_dir}{feature_name}.txt", 'r') as fp:
                    feature_pre = fp.read()
                feature_pre = feature_pre.split("\n")
                feature = [float(t) for t in feature_pre[:-1]]
                return feature

            torch.save(danceformer.state_dict(), f'{model_params_path}_epoch_{epoch}_batch_{i}.pt')

            #avg_prob_non_empty = open_txt_file("avg_prob_non_empty", avg_prob_non_empty)
            #avg_prob_non_empty_tmp = torch.tensor(avg_prob_non_empty)
            #torch.save(avg_prob_non_empty_tmp, f"{details_dir}avg_prob_non_empty_{i}_epoch_{epoch}")

            #if len(greater_ratio_values) != 0:
            #    # print(f"Greater values (ratio): {[t.item() for t in greater_ratio_values]}")
            #    greater_ratio_values_to_write = [str(t.item()) + '\n' for t in greater_ratio_values]
            #    torch.save(greater_ratio_values_to_write[0], f"{details_dir}greater_ratio_values_{i}_epoch_{epoch}")
            #else:
                # gute wahl??
            #    torch.save(0, f"{details_dir}greater_ratio_values_{i}_epoch_{epoch}")



    #token_probs_all = {k: [] for k in range(len(danceformer.indexed_vocabulary))}
    #for sublist in outputs:
    #    token_probs = {k: [] for k in range(len(danceformer.indexed_vocabulary))}
    #    for i, elem in enumerate(sublist):
    #        token_probs[i].append(sublist[i])

    #for i, elem in enumerate(token_probs):
    #    print(f"Token_probs {i}: {token_probs[i]}")
    print(f'confusion matrix (empty, non empty) \n{cm}')
    print(f'confusion matrix rt (right token, not right token) \n{cm_rt}')
    print(f'confusion matrix et_rt, where non empty token (right token, not right token) \n{cm_et_rt}')
    print(f'confusion matrix et_at, where on empty token (non empty token, empty token) \n{cm_et_at}')

    return cur_loss

# def evaluate(model, eval_data):
        #     total_vloss = 0.
        #     # Disable gradient computation and reduce memory consumption.
        #     with torch.no_grad():
        #           for i, vdata in enumerate(validation_loader):
        #               #load data
        #               data_imgs, diffs, charts = vdata
        #               data_imgs = data_imgs.to(device)
        #               diffs = diffs.to(device)
        #               charts = charts.to(device)
        #
        #               # sets pre_conv_imgs, diff_reps
        #               data_imgs, diffs = danceformer.forward(data_imgs, diffs)
        #               data_imgs = data_imgs.to(device)
        #               diffs = diffs.to(device)
        #               charts = charts.to(device)
        #               trans_model = danceformer.trans_model.to(device)
        #
        #               voutput = trans_model(data_imgs, diffs, charts).to(device)
        #               voutput_flat = voutput.view(-1, danceformer.ntokens).to(device)
        #               charts = charts[:, 1:]
        #               # charts_flat = charts.view(-1).to(device)
        #               charts_flat = charts.reshape(charts.shape[0] * charts.shape[1])
        #
        #               # charts[:, 1:]
        #               vloss = danceformer.criterion(voutput_flat, charts_flat)

        #               total_vloss += vloss
        #       # or return as in training per log_intervall, -1 right ?? oder i+1
        #       # (len(validation_loader) - 1)
        #     return total_vloss / (i+1)


def pad(x, length):
    # .cuda()?
    return torch.cat((x, torch.zeros((x.shape[0], x.shape[1], (length - x.shape[2])))), dim=2)


def custom_collate_fn(batch, max_len_specs, vocabulary, folder, specs_folder, device):
    # make list of each part (specs, difficulties, chart_elements)-> padded_batch per batch
    # maximum length given in seconds -> for chart: Pad each item to the maximum length

    specs = []
    difficulties = []
    chart_elements = []
    # makes sure all frames/spectrograms have same length, here min length, cut rest
    overall_min_frame_size = math.inf

    # look at RandomSubsequences
    for item in batch:
        song_path, waveform, sample_rate, song_difficulty, song_chart, song_mel_spec_23, song_mel_spec_46, \
        song_mel_spec_93 = item

        # calculate the duration: dividing the amount of elements of the audio by the samplerate
        length_song = waveform.shape[1] / sample_rate
        # in seconds
        start = torch.randint(int(length_song - max_len_specs), [1])
        sek_hop_length = (10 / 1000)

        s_ind = int(start / sek_hop_length)
        e_ind = int((start + max_len_specs) / sek_hop_length)

        # mel spectrogram transform
        specs_per_song = []

        sub_mel_spectogram_23 = song_mel_spec_23[:, :, s_ind:e_ind]
        sub_mel_spectogram_46 = song_mel_spec_46[:, :, s_ind:e_ind]
        sub_mel_spectogram_93 = song_mel_spec_93[:, :, s_ind:e_ind]

        specs_per_song.append(sub_mel_spectogram_23)
        specs_per_song.append(sub_mel_spectogram_46)
        specs_per_song.append(sub_mel_spectogram_93)

        # optional, save spectrogram snippets
        # initializing separator, for saving slice in right folder
        if specs_folder:
            separator = "/"
            result = song_path[:-4].rsplit(separator, 1)[0]
            result = result[len(folder):]
            path = f'{specs_folder}{result}'
            os.makedirs(path, exist_ok=True)
            torch.save(sub_mel_spectogram_23, f"{path}/mel_spectrogram_23.pt")
            torch.save(sub_mel_spectogram_46, f"{path}/mel_spectrogram_46.pt")
            torch.save(sub_mel_spectogram_93, f"{path}/mel_spectrogram_93.pt")

        # Slicing
        # frame_offset, num_frames = start * sample_rate, sample_rate * max_len_specs
        # print(f"Frame offset: {frame_offset}")
        # print(f"Number frames: {num_frames}")

        # unify previous spectrograms, maybe max, and fill ? or min and cut, now latter
        min_frame_size = math.inf
        for elem in specs_per_song:
            # for all previous saved specs, adjust size
            if elem.size()[2] < min_frame_size:
                min_frame_size = elem.size()[2]
                if min_frame_size < overall_min_frame_size:
                    overall_min_frame_size = min_frame_size
                    new_specs = []
                    # Problem: tensor not equal size, [1, 80, 5001] at entry 0 and [1, 80, 5000] at entry 2 ...
                    # update length of saved spectrograms, if too large (new ones shorter), shorten
                    # maybe change this to zero values, might get problem at border if cutted?
                    for spec in specs:
                        new_specs.append(spec[:, :, :overall_min_frame_size])
                    specs = new_specs
                else:
                    min_frame_size = overall_min_frame_size

        specs_unified = []

        # unify current song spectrograms
        for spec in specs_per_song:
            specs_unified.append(spec[:, :, :min_frame_size])

        # transform everything to tensors
        specs_unified = torch.stack(specs_unified)
        specs_unified = torch.squeeze(specs_unified, 1)
        specs.append(specs_unified.float())

        # charts: need to take appropriate window of chart
        # have intervals of 0.125 sec (range_dist) with criteria time_step > timed_note.time >= time_step - 0.125
        # interval ranges -> example [0,0.125)
        # we have a start token! +1 token always shifted
        range_dist = 1 / 64

        # take sublist of start to max_len_specs, need to transform start to window of chart
        # if chart not too short, take sublist of notes
        # start in seconds /range dist same as *64, gives right interval, but +1 (start token, always shifted)
        start_time = int(start / range_dist) + 1
        range_time = max_len_specs / range_dist
        range_end = int(start_time + range_time)
        # end out of range, because chart already ended...
        if range_end > len(song_chart):
            # take end chart token
            end_token_tmp = song_chart[-1]
            song_chart = song_chart[:-1]
            # add empty chart token
            while range_end != len(song_chart)-1:
                song_chart.append(indexed_vocabulary.get("0000"))
            song_chart.append(end_token_tmp)

        chart_elements.append(song_chart[start_time:range_end])
        difficulties.append(torch.tensor(int(song_difficulty)))

    # make tensors
    specs = torch.stack(specs, dim=0)
    specs = torch.squeeze(specs, 1)

    # make tensors
    difficulties = torch.stack(difficulties)
    # was macht das? -> neue dimension (1)
    # difficulties = difficulties[:, None]

    new_chart_elements = torch.stack([torch.tensor(cha) for cha in chart_elements])
    # Create a list to store the padded tensors, look here might want tensors?

    padded_batch = []
    padded_batch.append(specs)
    padded_batch.append(difficulties)
    padded_batch.append(new_chart_elements)
    # torch.stack(padded_batch)
    return padded_batch


def setup_parser():
    out = argparse.ArgumentParser()
    # # cluster version
    folder_given = "approach_empty_non_empty"
    save_given = "approach_empty_non_empty_new"
    out.add_argument('--get_vocabulary', default=f'/scratch/grzonkow/{folder_given}/vocabulary_empty_non_empty_approach_latest.pkl', type=str, help="Path to load vocabulary")
    out.add_argument('--save_vocabulary', default=f'/scratch/grzonkow/{save_given}/vocabulary_new_{folder_given}.pkl',
                     type=str, help="Path to safe vocabulary")

    # if built_Dataset, dataset does not matter
    out.add_argument('--built_dataset', default=f'/scratch/grzonkow/output', type=str, help="Path to dataset folder")
    out.add_argument('--built_dataset_charts', default=f'/scratch/grzonkow/{folder_given}/output', type=str, help="Path to dataset folder")
    out.add_argument('--dataset', default=None, type=str, help="Path to dataset")
    out.add_argument('--save_dataset', default=f'/scratch/grzonkow/{save_given}/train_dataset_{folder_given}.pkl',
                     type=str, help="Path to safe dataset")
    out.add_argument('--old_version', default=False, type=str,
                     help="Indicates if old version without start end tokens need to be loaded")
    out.add_argument('--old_version_voc', default=False, type=str,
                     help="Indicates if old version without start end tokens need to be loaded")

    out.add_argument('--folder', default=r'/work/MLShare/StepMania/data/cleaned/allowed_meter_difference_of_2/',
                     type=str, help="Path to folder containing packages with music and sm files")
    out.add_argument('--batch_size', default=100, type=int, help="Size of a batch")
    out.add_argument('--number_epochs', default=3, type=int, help="Number of epochs for training")

    out.add_argument('--model_params_path',
                     default=f'/scratch/grzonkow/{save_given}/models/model_params_2702_100_batch_size',
                     type=str, help="Path to safe model parameters")
    out.add_argument('--load_model_params_path', default=None, type=str, help="Path to load model parameters")
    out.add_argument('--save_dir_models', default=f"/scratch/grzonkow/{save_given}/models",
                     type=str, help="Path to save model files")

    out.add_argument('--save_dir', default=f"/scratch/grzonkow/{save_given}/output",
                     type=str, help="Path to safe dataset points")
    # if you want to save the snippets of the melspectograms
    out.add_argument('--specs_folder', default=None,
                     type=str, help="Path to save melspecs snippets")
    out.add_argument('--details_dir', default=f"/scratch/grzonkow/{save_given}/details/",
                     type=str, help="Path to save details")
    out.add_argument('--model_params_path_nan_problem', default=None,
                     type=str, help="Path to save problems")





    # folder_given = "Model_naive_approach"
    # out.add_argument('--get_vocabulary', default=f'/scratch/grzonkow/{folder_given}/vocabulary_{folder_given}.pkl', type=str, help="Path to load vocabulary")
    # out.add_argument('--save_vocabulary', default=f'/scratch/grzonkow/{folder_given}/vocabulary_new_{folder_given}.pkl',
    #                  type=str, help="Path to safe vocabulary")
    #
    # # if built_Dataset, dataset does not matter
    # out.add_argument('--built_dataset', default=f'/scratch/grzonkow/output/', type=str, help="Path to dataset folder")
    # out.add_argument('--built_dataset_charts', default=f'/scratch/grzonkow/{folder_given}/output/', type=str, help="Path to dataset folder")
    # out.add_argument('--dataset', default=None, type=str, help="Path to dataset")
    # out.add_argument('--save_dataset', default=f'/scratch/grzonkow/{folder_given}/train_dataset_{folder_given}.pkl',
    #                  type=str, help="Path to safe dataset")
    # out.add_argument('--old_version', default=False, type=str,
    #                  help="Indicates if old version without start end tokens need to be loaded")
    # out.add_argument('--old_version_voc', default=True, type=str,
    #                  help="Indicates if old version without start end tokens need to be loaded")
    #
    # out.add_argument('--folder', default=r'/work/MLShare/StepMania/data/cleaned/allowed_meter_difference_of_2/',
    #                  type=str, help="Path to folder containing packages with music and sm files")
    # out.add_argument('--batch_size', default=100, type=int, help="Size of a batch")
    # out.add_argument('--number_epochs', default=3, type=int, help="Number of epochs for training")
    #
    # out.add_argument('--model_params_path',
    #                  default=f'/scratch/grzonkow/{folder_given}/models/model_params_2702_100_batch_size',
    #                  type=str, help="Path to safe model parameters")
    # out.add_argument('--load_model_params_path', default=None, type=str, help="Path to load model parameters")
    # out.add_argument('--save_dir_models', default=f"/scratch/grzonkow/{folder_given}/models",
    #                  type=str, help="Path to save model files")
    #
    # out.add_argument('--save_dir', default=f"/scratch/grzonkow/{folder_given}/output",
    #                  type=str, help="Path to safe dataset points")
    # # if you want to save the snippets of the melspectograms
    # out.add_argument('--specs_folder', default=None,
    #                  type=str, help="Path to save melspecs snippets")
    # out.add_argument('--details_dir', default=f"/scratch/grzonkow/{folder_given}/details/",
    #                  type=str, help="Path to save details")
    # out.add_argument('--model_params_path_nan_problem', default=f"/scratch/grzonkow/{folder_given}/nan_problem/",
    #                  type=str, help="Path to save problems")




    #local version
    # folder_given = "naive_approach"
    #
    # #out.add_argument('--get_vocabulary', default=None,type=str, help="Path to load vocabulary")
    # out.add_argument('--get_vocabulary',
    #                  default=f'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/vocabulary.pkl',
    #                   type=str, help="Path to load vocabulary")
    # out.add_argument('--save_vocabulary', default=f'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/vocabulary.pkl',
    #                  type=str, help="Path to save vocabulary")
    #
    # # if built_Dataset, dataset does not matter
    # out.add_argument('--built_dataset', default=f'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/output/', type=str, help="Path to dataset folder")
    # out.add_argument('--dataset', default=None, type=str, help="Path to dataset")
    # out.add_argument('--save_dataset',
    #                  default=f'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/train_dataset_latest.pkl',
    #                  type=str, help="Path to safe dataset")
    #
    # out.add_argument('--folder', default=r'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_pack/',
    #                  type=str, help="Path to folder containing packages with music and sm files")
    # out.add_argument('--batch_size', default=2, type=int, help="Size of a batch")
    # out.add_argument('--number_epochs', default=3, type=int, help="Number of epochs for training")
    #
    # out.add_argument('--model_params_path',
    #                  default=f'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/model_params',
    #                  type=str, help="Path to safe model parameters")
    # out.add_argument('--load_model_params_path', default=None, type=str, help="Path to load model parameters")
    # #out.add_argument('--load_model_params_path', default=r'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/new model/model_params_2302.pt',
    # #                 type=str, help="Path to load model parameters")
    # out.add_argument('--save_dir_models',
    #                  default=f"C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/models/",
    #                  type=str, help="Path to safe dataset points")
    #
    # #out.add_argument('--save_dir', default=None, type=str, help="Path to dataset")
    # out.add_argument('--save_dir', default=f"C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/output",
    #                  type=str, help="Path to safe dataset points")
    # out.add_argument('--specs_folder', default=None,
    #                  type=str, help="Path to save melspecs snippets")
    # out.add_argument('--details_dir', default=f"C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/details/",
    #                   type=str, help="Path to safe dataset points")
    # out.add_argument('--model_params_path_nan_problem', default=f"C:/Users/cassi/OneDrive/Desktop/Master_Thesis/{folder_given}/nan_problem/",
    #                  type=str, help="Path to safe problems")
    # #

    # #local version
    #out.add_argument('--get_vocabulary',
    #                 default=None,
    #                 type=str, help="Path to load vocabulary")
    # out.add_argument('--get_vocabulary', default=r'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/cluster/vocabulary.pkl',
    #                  type=str, help="Path to load vocabulary")
    # out.add_argument('--save_vocabulary', default=r'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/vocabulary_new.pkl',
    #                  type=str, help="Path to safe vocabulary")
    # #out.add_argument('--dataset', default="C:/Users/cassi/OneDrive/Desktop/Master_Thesis/train_dataset_latest.pkl",
    # #                type=str, help="Path to dataset")
    # out.add_argument('--old_version',
    #                  default=False,
    #                  type=str, help="Indicates if old version without start end tokens need to be loaded")
    # out.add_argument('--dataset', default='C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/train_dataset_latest.pkl', type=str, help="Path to dataset")
    # out.add_argument('--built_dataset', default=None, type=str, help="Path to dataset folder")
    # #out.add_argument('--built_dataset', default="C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/output", type=str, help="Path to dataset folder")
    # out.add_argument('--save_dataset',
    #                  default='C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/train_dataset_latest.pkl',
    #                  type=str, help="Path to safe dataset")
    # out.add_argument('--folder', default=r'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/test_pack/',
    #                  type=str, help="Path to folder containing packages with music and sm files")
    # out.add_argument('--model_params_path',
    #                  default=r'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/model_params',
    #                  type=str, help="Path to safe model parameters")
    # out.add_argument('--load_model_params_path', default=None, type=str, help="Path to load model parameters")
    # #out.add_argument('--load_model_params_path', default=r'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/new model/model_params_2302.pt',
    # #                 type=str, help="Path to load model parameters")
    # #out.add_argument('--save_dir', default=None, type=str, help="Path to dataset")
    # out.add_argument('--save_dir', default="C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/output",
    #                  type=str, help="Path to safe dataset points")
    # out.add_argument('--specs_folder', default="C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/copy_dataset/",
    #                  type=str, help="Path to safe dataset points")
    # out.add_argument('--details_dir', default="C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/",
    #                   type=str, help="Path to safe dataset points")
    # out.add_argument('--model_params_path_nan_problem', default="C:/Users/cassi/OneDrive/Desktop/Master_Thesis/test_cluster/nan_problem/",
    #                  type=str, help="Path to safe problems")

    return out


if __name__ == '__main__':
    parser = setup_parser()
    args, unknown = parser.parse_known_args()
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
    # create some folders
    # os.makedirs(args.model_params_path_nan_problem, exist_ok=True)
    os.makedirs(args.details_dir, exist_ok=True)
    os.makedirs(args.save_dir_models, exist_ok=True)

    # some paths
    #best_model_params_path = os.path.join(tempdir, "best_model_params.pt")
    model_params_path = args.model_params_path
    details_dir = args.details_dir

    # does this make sense at all? for later on: delete, might be good for debugging
    random.seed(10)

    # print info about torch
    print(torch.cuda.is_available())
    print(torch.__version__)
    print(torch.__path__)
    print(torch.version.cuda)
    print(torch.backends.cudnn.version())

    # prepare pack/folder of all songs
    folder = args.folder
    pkgs = os.listdir(folder)
    songs = []
    for pkg in pkgs:
        song_folder_names = os.listdir(f"{folder}{pkg}")
        for song in song_folder_names:
            songs.append(f"{folder}{pkg}/{song}")

    # preprocessing, vocabulary
    if args.get_vocabulary is None:
        indexed_vocabulary = built_chart_vocabulary(songs)
        if args.old_version_voc:
            # need to add start and endtoken
            # Get the last value entry or len(indexed_vocabulary)-1)
            last_value = list(indexed_vocabulary.values())[-1]
            start_token = "<s>"
            end_token = "<e>"
            indexed_vocabulary[start_token] = last_value+1
            indexed_vocabulary[end_token] = last_value+2
            print(f"With start and end token: \n{indexed_vocabulary}")
            print(f"Length updated vocabulary: \n{len(indexed_vocabulary)}")
            # save vocabulary for reusing
        with open(args.save_vocabulary, 'wb') as fp:
            pickle.dump(indexed_vocabulary, fp)
            print('dictionary saved successfully to file')
    else:
        # Read dictionary pkl file
        with open(f'{args.get_vocabulary}', 'rb') as fp:
            indexed_vocabulary = pickle.load(fp)
            print('Loaded dictionary:')
            print(indexed_vocabulary)
        # only once, if old vocab
        if args.old_version_voc:
            # Get the last value entry or len(indexed_vocabulary)-1)
            last_value = list(indexed_vocabulary.values())[-1]
            start_token = "<s>"
            end_token = "<e>"
            indexed_vocabulary[start_token] = last_value + 1
            indexed_vocabulary[end_token] = last_value + 2
            print(f"With start and end token: \n{indexed_vocabulary}")
            print(f"Length updated vocabulary: \n{len(indexed_vocabulary)}")
            # save vocabulary for reusing
            with open(args.save_vocabulary, 'wb') as fp:
                pickle.dump(indexed_vocabulary, fp)
                print('dictionary saved successfully to file')
    print("--------------------------------------------------------------------------------- \n")

    # initializes MusicNet, MlP/DifficultyNet, Transformer
    danceformer = Danceformer(indexed_vocabulary).to(device)
    # print information about danceformer
    print(danceformer)
    print("Number parameters:")
    print(sum(p.numel() for p in danceformer.parameters() if p.requires_grad))

    # load pretrained model if given
    if args.load_model_params_path:
        danceformer.load_state_dict(torch.load(args.load_model_params_path, map_location=torch.device('cpu')))
        print("Model loaded")

    start_time_dataset_step = time.time()

    # if given, just built dataset given the output folder path
    if args.built_dataset:
        # if old version: added start and end tokens
        if args.old_version:
            # alternativ value des vorletzten und letzten vocab eintrags
            start_token = indexed_vocabulary.get("<s>")
            end_token = indexed_vocabulary.get("<e>")

            # took abot 50 minutes for cluster data
            path_to_charts = f'{args.built_dataset}/all_charts/'
            for idx in range(0, len(os.listdir(path_to_charts))):
                # create new chart with start token, extend it with old chart, append end token to it
                with open(f"{path_to_charts}{idx}.pkl", 'rb') as fp:
                    new_song_chart = []
                    new_song_chart.append(start_token)
                    song_chart = pickle.load(fp)
                    new_song_chart.extend(song_chart)
                    # update with start and end token
                    new_song_chart.append(end_token)
                # save again, old charts with start and end tokens
                with open(f"{path_to_charts}{idx}.pkl", 'wb') as fp:
                    pickle.dump(new_song_chart, fp)

            # test if it worked for one file
            # with open(f"{path_to_charts}0.pkl", 'rb') as fp:
            #    song_chart = pickle.load(fp)
            #   # see start token
            #    print(song_chart)
            #   # see end token
            #    print(song_chart[-1])

        # in ms
        win_length_list_ms = [23, 46, 93]

        # Data loading.
        train_dataset = SongPacksDataset(
            f'{args.built_dataset}/all_paths/',
            f'{args.built_dataset}/all_waves/',
            f'{args.built_dataset}/all_sample_rates/',
            f'{args.built_dataset}/all_difficulties/',
            f'{args.built_dataset_charts}/all_charts/',
            f"{args.built_dataset}/mel_specs_{win_length_list_ms[0]}/",
            f"{args.built_dataset}/mel_specs_{win_length_list_ms[1]}/",
            f"{args.built_dataset}/mel_specs_{win_length_list_ms[2]}/",
            device,
            indexed_vocabulary,
        )

        # really needed?, already same data in path built_dataset
        with open(args.save_dataset, 'wb') as output:
            pickle.dump(train_dataset, output)

    else:
        # nothing there, create dataset
        if args.dataset is None:
            # create folders
            os.makedirs(f"{args.save_dir}/all_paths/", exist_ok=True)
            os.makedirs(f"{args.save_dir}/all_waves/", exist_ok=True)
            os.makedirs(f"{args.save_dir}/all_sample_rates/", exist_ok=True)
            os.makedirs(f"{args.save_dir}/all_difficulties/", exist_ok=True)
            os.makedirs(f"{args.save_dir}/all_charts/", exist_ok=True)
            # in ms
            win_length_list_ms = [23, 46, 93]
            for win in win_length_list_ms:
                os.makedirs(f"{args.save_dir}/mel_specs_{win}/", exist_ok=True)

            get_paths_diff_charts(args.save_dir, songs, indexed_vocabulary, device)
            elapsed_dataset_step = time.time() - start_time_dataset_step
            print('-' * 89)
            print(f'Dataset created, time: {elapsed_dataset_step:5.2f}s')
            print('-' * 89)

            # Data loading.
            train_dataset = SongPacksDataset(
                f'{args.save_dir}/all_paths/',
                f'{args.save_dir}/all_waves/',
                f'{args.save_dir}/all_sample_rates/',
                f'{args.save_dir}/all_difficulties/',
                f'{args.save_dir}/all_charts/',
                f"{args.save_dir}/mel_specs_{win_length_list_ms[0]}/",
                f"{args.save_dir}/mel_specs_{win_length_list_ms[1]}/",
                f"{args.save_dir}/mel_specs_{win_length_list_ms[2]}/",
                device,
            )

            # save dataset
            with open(args.save_dataset, 'wb') as output:
                pickle.dump(train_dataset, output)

        # dataset completely given as file
        else:
            with open(args.dataset, 'rb') as data:
                train_dataset = pickle.load(data)
                elapsed_dataset_step = time.time() - start_time_dataset_step
                print('-' * 89)
                print(f'Dataset loaded, elapsed time: {str(timedelta(seconds=elapsed_dataset_step))}')
                print('-' * 89)


    # check batch, output, maybe change to class / transform !
    start_time_train_loader_step = time.time()
    train_loader = torch.utils.data.DataLoader(
        train_dataset, batch_size=args.batch_size, shuffle=True, num_workers=0,
        collate_fn=partial(custom_collate_fn, max_len_specs=5, vocabulary=indexed_vocabulary, folder=folder,
                           specs_folder=args.specs_folder, device=device))
    elapsed_train_loader_step = time.time() - start_time_train_loader_step
    print('-' * 89)
    print(f'Data loaded, time: {elapsed_train_loader_step:5.2f}s')
    print(f'Length train loader: {len(train_loader)}')
    print('-' * 89)

    epochs = args.number_epochs
    # timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
    # best_vloss = float('inf')
    # per epoch, save median values of different features
    median_times = []
    median_losses = []
    median_losses_v_2 = []
    median_ppls = []
    median_accuracies = []
    #median_non_empty_tokens = []
    #median_right_tokens = []
    #median_avg_right_tokens = []
    median_ppls_v_2 = []
    median_accuracies_v_2 = []
    median_accuracies_v_3 = []

    # https://pytorch.org/tutorials/beginner/transformer_tutorial.html
    # with TemporaryDirectory() as tempdir:
    #     best_model_params_path = os.path.join(tempdir, "best_model_params.pt")

    for epoch in range(1, epochs + 1):
        print(f'EPOCH {epoch}:')
        epoch_start_time = time.time()

        danceformer.trans_model.train().to(device)  # turn on train mode
        # create new files for features: write values for each epoch
        #if os.path.exists(f"{details_dir}avg_prob_non_empty.txt"):
        #    os.remove(f"{details_dir}avg_prob_non_empty.txt")
        if os.path.exists(f"{details_dir}ms_per_batch.txt"):
            os.remove(f"{details_dir}ms_per_batch.txt")
        if os.path.exists(f"{details_dir}ppl.txt"):
            os.remove(f"{details_dir}ppl.txt")
        if os.path.exists(f"{details_dir}ppl_v_2.txt"):
            os.remove(f"{details_dir}ppl_v_2.txt")
        if os.path.exists(f"{details_dir}loss.txt"):
            os.remove(f"{details_dir}loss.txt")
        if os.path.exists(f"{details_dir}loss_v_2.txt"):
            os.remove(f"{details_dir}loss_v_2.txt")
        if os.path.exists(f"{details_dir}accuracy.txt"):
            os.remove(f"{details_dir}accuracy.txt")
        if os.path.exists(f"{details_dir}accuracy_v_2.txt"):
            os.remove(f"{details_dir}accuracy_v_2.txt")
        if os.path.exists(f"{details_dir}accuracy_v_3.txt"):
            os.remove(f"{details_dir}accuracy_v_3.txt")

        #if os.path.exists(f"{details_dir}greater_ratio_values.txt"):
        #    os.remove(f"{details_dir}greater_ratio_values.txt")
        #if os.path.exists(f"{details_dir}avg_prob_non_empty.txt"):
        #    os.remove(f"{details_dir}avg_prob_non_empty.txt")
        #if os.path.exists(f"{details_dir}average_right_tokens.txt"):
        #    os.remove(f"{details_dir}average_right_tokens.txt")

        avg_loss = train_one_epoch(danceformer, train_loader, epoch, model_params_path, device, details_dir,
                            args.model_params_path_nan_problem)

        # https://pytorch.org/tutorials/beginner/introyt/trainingyt.html
        # Set the model to evaluation mode, disabling dropout and using population
        # statistics for batch normalization.
        # in beginning split in train and val dataset
        # as train dataset, load val_dataset and val_dataloader, then:
        # danceformer.trans_model.eval().to(device)
        # #avg_val_loss
        # val_loss = evaluate(model, val_data)
        # val_ppl = math.exp(val_loss)
        # elapsed = time.time() - epoch_start_time
        # print('-' * 89)
        # Print(f'| end of epoch {epoch:3d} | time: {elapsed:5.2f}s | '
        #        f'valid loss {val_loss:5.2f} | valid ppl {val_ppl:8.2f}')
        # print('-' * 89)
        # if val_loss < best_val_loss:
        #             best_val_loss = val_loss
        #             torch.save(model.state_dict(), best_model_params_path)
        elapsed = time.time() - epoch_start_time
        # save model per epoch
        torch.save(danceformer.state_dict(), f'{model_params_path}_epoch_{epoch}.pt')
        # print epoch that ended, elapsed time
        print('-' * 89)
        print(f'| end of epoch {epoch:3d} | time: {elapsed:5.2f}s | ')


        # f'valid loss {val_loss:5.2f} | valid ppl {val_ppl:8.2f}')

        def open_txt_file(feature_name):
            with open(f"{details_dir}{feature_name}.txt", 'r') as fp:
                feature_pre = fp.read()
            feature_pre = feature_pre.split("\n")
            feature_name = [float(t) for t in feature_pre[:-1] if t != '']
            return feature_name

        # load written data for different features and transform them in correct format
        #avg_prob_non_empty = open_txt_file("avg_prob_non_empty")
        ms_per_batch = open_txt_file("ms_per_batch")
        ppl = open_txt_file("ppl")
        ppl_v_2 = open_txt_file("ppl_v_2")
        loss = open_txt_file("loss")
        loss_v_2 = open_txt_file("loss_v_2")
        accuracy = open_txt_file("accuracy")
        accuracy_v_2 = open_txt_file("accuracy_v_2")
        accuracy_v_3 = open_txt_file("accuracy_v_3")

        #average_right_tokens = open_txt_file("average_right_tokens")


        # special case, we have list of ratios per epoch, empty list and filled ones
        #with open(f"{details_dir}greater_ratio_values.txt", 'r') as fp:
        #    feature_pre = fp.read()
        #feature_pre = feature_pre.split("\n")
        #greater_ratio_values_all = []
        #for t in feature_pre[:-1]:
        #    if t != str([]):
        #        greater_ratio_values_all.extend(float(elem) for elem in t.strip('][').split(', '))
        #greater_ratio_values = greater_ratio_values_all



        # print status of features
        print(f"Last loss: {avg_loss}")
        median_time = sum(ms_per_batch)/len(ms_per_batch)
        median_times.append(median_time)
        print(f"Median time in ms per batch: {median_time}")
        median_ppl = sum(ppl)/len(ppl)
        median_ppls.append(median_ppl)
        print(f"Median ppl per batch: {median_ppl}")
        median_loss = sum(loss)/len(loss)
        median_losses.append(median_loss)
        print(f"Median loss per batch: {median_loss}")
        median_accuracy = sum(accuracy) / len(accuracy)
        median_accuracies.append(median_accuracy)
        print(f"Median accuracy per batch: {median_accuracy}")
        median_loss_v_2 = sum(loss_v_2) / len(loss_v_2)
        median_losses_v_2.append(median_loss_v_2)
        print(f"Median loss v2 per batch: {median_loss_v_2}")
        median_accuracy_v_2 = sum(accuracy_v_2) / len(accuracy_v_2)
        median_accuracies_v_2.append(median_accuracy_v_2)
        print(f"Median accuracy v_2 per batch: {median_accuracy_v_2}")
        median_accuracy_v_3 = sum(accuracy_v_3) / len(accuracy_v_3)
        median_accuracies_v_3.append(median_accuracy_v_3)
        print(f"Median accuracy v_3 per batch: {median_accuracy_v_3}")
        median_ppl_v_2 = sum(ppl_v_2) / len(ppl_v_2)
        median_ppls_v_2.append(median_ppl_v_2)
        print(f"Median ppl v2 per batch: {median_ppl_v_2}")

        #median_non_empty_token = sum(avg_prob_non_empty)/len(avg_prob_non_empty)
        #median_non_empty_tokens.append(median_non_empty_token)
        #print(f"Median non empty tokens: {median_non_empty_token}")
        # special case, no values since not saved per batch
        #if len(greater_ratio_values) != 0:
        #    median_right_token = sum(greater_ratio_values) / len(greater_ratio_values)
        #    median_right_tokens.append(median_right_token)
        #    print(f"Median right tokens: {median_right_token}")
        #median_average_right_token = sum(average_right_tokens)/len(average_right_tokens)
        #print(f"Median average right tokens: {median_average_right_token}")
        #median_non_empty_tokens.append(median_average_right_token)
        print('-' * 89)

        # transform features to tensors
        ppl = torch.tensor(ppl)
        ppl_v_2 = torch.tensor(ppl_v_2)
        ms_per_batch = torch.tensor(ms_per_batch)
        loss = torch.tensor(loss)
        accuracy = torch.tensor(accuracy)
        loss_v_2 = torch.tensor(loss_v_2)
        accuracy_v_2 = torch.tensor(accuracy_v_2)
        accuracy_v_3 = torch.tensor(accuracy_v_3)

        #avg_prob_non_empty = torch.tensor(avg_prob_non_empty)
        #average_right_tokens = torch.tensor(average_right_tokens)
        #greater_ratio_values = torch.tensor(greater_ratio_values)

        # save values per log_interval (for 1: batch) for given epoch
        torch.save(ppl, f"{details_dir}ppl_batches_epoch_{epoch}")
        torch.save(ppl_v_2, f"{details_dir}ppl_v_2_batches_epoch_{epoch}")
        torch.save(ms_per_batch, f"{details_dir}ms_all_batches_epoch_{epoch}")
        torch.save(loss, f"{details_dir}loss_epoch_{epoch}")
        torch.save(accuracy, f"{details_dir}accuracy_epoch_{epoch}")
        torch.save(loss_v_2, f"{details_dir}loss_v_2_epoch_{epoch}")
        torch.save(accuracy_v_2, f"{details_dir}accuracy_v_2_epoch_{epoch}")
        torch.save(accuracy_v_3, f"{details_dir}accuracy_v_3_epoch_{epoch}")

        #torch.save(avg_prob_non_empty, f"{details_dir}avg_prob_non_empty_epoch_{epoch}")
        #torch.save(greater_ratio_values, f"{details_dir}greater_ratio_values_per_batch_epoch_{epoch}")



    # print after all epochs, overview: median feature values per epoch for all epochs
    print(f"Median times in ms per epoch: {median_times}")
    print(f"Median ppl per epoch: {median_ppls}")
    print(f"Median ppl v2 per epoch: {median_ppl_v_2}")
    print(f"Median loss per epoch: {median_losses}")
    print(f"Median loss v2 per epoch: {median_losses_v_2}")
    print(f"Median accuracy per epoch: {median_accuracies}")
    print(f"Median accuracy v2 per epoch: {median_accuracies_v_2}")
    print(f"Median accuracy v3 per epoch: {median_accuracies_v_3}")

    #print(f"Median non empty tokens per epoch: {median_non_empty_tokens}")
    #print(f"Median right tokens per epoch: {median_right_tokens}")
    # danceformer.trans_model.load_state_dict(torch.load(best_model_params_path))  # load best model states
    # Detach tensors from GPU
    # plot_loss_accuracy(train_loss, train_acc, validation_loss, validation_acc)