import math
import os
import pickle
import torch
import torch.nn.functional as F
import simfile
from preliminary_experiment import *
from torch import nn, Tensor
from torch.nn import TransformerEncoder, TransformerEncoderLayer, TransformerDecoderLayer, TransformerDecoder
from collections import OrderedDict
# maybe graphviz
# maybe torchview, draw_graph

# from https://deeplearning.neuromatch.io/projects/ComputerVision/spectrogram_analysis.html
# architecture as in https://github.com/chrisdonahue/ddc
# @inproceedings{donahue2017dance,
#   title={Dance Dance Convolution},
#   author={Donahue, Chris and Lipton, Zachary C and McAuley, Julian},
#   booktitle={Proceedings of the 34th International Conference on Machine Learning},
#   year={2017},
# }
class Music_Net(nn.Module):

    def __init__(self) -> object:
        """Intitalize neural net layers"""
        super(Music_Net, self).__init__()
        # conv1d, original paper? freq as channel
        # better: no padding, use image better
        # padding input, check before if needed?
        #self.conv1 = nn.Conv2d(in_channels=3, out_channels=10, kernel_size=(7, 3), stride=(1, 3), padding=(3, 0))
        #self.conv2 = nn.Conv2d(in_channels=10, out_channels=20, kernel_size=(3, 3), stride=(1, 3), padding=(1, 0))
        #self.m = nn.MaxPool2d(kernel_size=(1, 3), stride=(1, 3))
        self.conv1 = nn.Conv1d(in_channels=240, out_channels=120, kernel_size=3, stride=1, padding=1)
        self.conv2 = nn.Conv1d(in_channels=120, out_channels=60, kernel_size=3, stride=1, padding=1)

        # self.batchnorm1 = nn.BatchNorm2d(num_features=8)
        # self.batchnorm2 = nn.BatchNorm2d(num_features=16)
        self.dropout = nn.Dropout(p=0.5, inplace=False)

    def forward(self, x):
        # need format: (batch), time_steps, bands, channels
        #print(f"CNN --------------------------")
        #print(f"Begin: {x.shape}")
        #batch, channels, bands, time_steps = x.shape
        #x = torch.reshape(x, [batch, time_steps, bands, channels])
        batch, channel, frequency, time = x.shape
        x = x.view(batch, -1, time)
        # Conv layer 1.
        x = self.conv1(x)
        #print(f"After 1.conv: {x.shape}")
        # x = self.batchnorm1(x)
        x = F.relu(x)
        #print(f"After 1.relu: {x.shape}")
        # only in frequency domain?
        # x = F.max_pool1d(x, kernel_size=3, stride=3)
        #x = self.m(x)
        #print(f"After first maxpool: {x.shape}")

        # Conv layer 2.
        x = self.conv2(x)
        #print(f"After 2.conv: {x.shape}")
        # x = self.batchnorm2(x)
        x = F.relu(x)
        # x = F.max_pool1d(x, kernel_size=3, stride=3)
        #print(f"After 2.relu: {x.shape}")
        #x = self.m(x)
        #print(f"After 2.maxpool: {x.shape}")
        # batch, time, freq, chanel dim check
        #x = x.view(-1, x.shape[-3], x.shape[-2]*x.shape[-1])
        #torch.Size([1, 20, 13840])
        #print("MusicNet output shape:")
        #print(x.shape)

        return x


class Difficulty_Net(nn.Module):
    def __init__(self):
        super().__init__()
        # https://machinelearningmastery.com/building-multilayer-perceptron-models-in-pytorch/
        self.mlp_model = nn.Sequential(OrderedDict([
            # first value is number of min charts... must be fixed or call per difficulty
            ('dense1', nn.Linear(1, 128)),
            ('act1', nn.ReLU()),
            ('dense2', nn.Linear(128, 128)),
            ('act2', nn.ReLU()),
            ('output', nn.Linear(128, 60)), #128,128
            # ('outact', nn.Sigmoid()),
        ]))

    def forward(self, difficulties):
        #print(f"Difficulties shape: {len(difficulties)}")
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        #difficulties = self.mlp_model(torch.tensor(difficulties.unsqueeze(1), dtype=torch.float32, device=device))
        difficulties = self.mlp_model(difficulties.unsqueeze(1).clone().detach().to(torch.float32).to(device))
        #difficulties = self.mlp_model((difficulties.unsqueeze(1)).to(torch.float32))
        #print("DifficultyNet output shape:")
        #print(difficulties.shape)

        return difficulties


# # from https://pytorch.org/tutorials/beginner/transformer_tutorial.html
class TransformerModel(nn.Module):

    def __init__(self, ntoken: int, d_model: int, nhead: int, d_hid: int,
                 nlayers: int, dropout: float = 0.5):
        super().__init__()
        self.model_type = 'Transformer'
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        self.embedding = nn.Embedding(ntoken, d_model).to(device)
        #self.embedding_tgt = nn.Embedding(ntoken_tgt, d_model)
        self.pos_encoder = PositionalEncoding(d_model, dropout).to(device)
        encoder_layers = TransformerEncoderLayer(d_model, nhead, d_hid, dropout, batch_first=True).to(device)
        decoder_layers = TransformerDecoderLayer(d_model, nhead, d_hid, dropout, batch_first=True).to(device)
        self.transformer_encoder = TransformerEncoder(encoder_layers, nlayers).to(device)
        self.transformer_decoder = TransformerDecoder(decoder_layers, nlayers).to(device)

        self.d_model = d_model
        self.linear = nn.Linear(d_model, ntoken).to(device)

        self.init_weights()

    def init_weights(self) -> None:
        initrange = 0.1
        self.embedding.weight.data.uniform_(-initrange, initrange)
        self.linear.bias.data.zero_()
        self.linear.weight.data.uniform_(-initrange, initrange)

    # trans_model(data_imgs, diffs, charts)
    def forward(self, data_imgs: Tensor, diffs: Tensor, charts: Tensor, src_mask: Tensor = None) -> Tensor:
        """
        Arguments:
            src: Tensor, shape ``[seq_len, batch_size]``
            src_mask: Tensor, shape ``[seq_len, seq_len]``

        Returns:
            output Tensor of shape ``[seq_len, batch_size, ntoken]``
        """
        # example torch.Size([2, 20, 480])    original given into music net: torch.Size([2, 3, 80, 500])
        #print(f"Initial data images shape: {data_imgs.shape}")
        # 64-bit integer (signed). torch.int64 or torch.long
        #data_imgs = torch.tensor(data_imgs).to(torch.long)

        # example torch.Size([2, 20])
        #print(f"Initial diffs shape: {diffs.shape}")

        # batch of difficulty combined with batch of images ...
        # new dimension
        # diffs = diffs.unsqueeze(0)
        # print(f"diffs shape: {diffs.shape}")
        # repeat entry in new dimension for number of frames (3)
        # diffs = diffs.repeat(data_imgs.shape[1], 1)
        # print(f"Adjusted diffs shape: {diffs.shape}")
        # diffs = torch.tensor(diffs).to(torch.long)
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

        # example torch.Size([2, 320])
        #print(f"Initial charts shape: {charts.shape}")

        # what happens here??
        #diffs = self.embedding(diffs) * math.sqrt(self.d_model)
        #print(f"src shape after embedding: {src.shape}")
        #print(f"diffs shape after embedding: {diffs.shape}")
        data_imgs = data_imgs.permute(0, 2, 1)
        data_imgs = self.pos_encoder(data_imgs).to(device)
        #print(f"Data images shape after pos encoder: {data_imgs.shape}")

        # here nan
        output = self.transformer_encoder(data_imgs).to(device)
        #print(f"Output transformer encoder: {output}")
        #print(f"Output transformer encoder shape: {output.shape}")

        #print(f"Diffs input shape: {diffs.unsqueeze(1).shape}")
        comb = torch.cat((output, diffs.unsqueeze(1)), 1).to(device)
        #print(f"comb shape after concatenating src with diffs: {comb.shape}")

        tgt_mask = None

        if tgt_mask is None:
            """Generate a square causal mask for the sequence. The masked positions are filled with float('-inf').
            Unmasked positions are filled with float(0.0).
            """
            tgt_mask = nn.Transformer.generate_square_subsequent_mask(charts.shape[1]-1).to(device)

        # out = transformer_decoder(tgt, memory),
        # tgt (Tensor) – the sequence to the decoder (required).
        # memory (Tensor) – the sequence from the last layer of the encoder (required).
        #out = self.transformer_decoder(comb, comb)
        #print(f"Dtype charts: {charts.dtype}")
        #charts = charts.to(torch.float32)
        charts = self.embedding(charts).to(device)
        charts = self.pos_encoder(charts).to(device)
        charts = charts.squeeze(0)
        #print(f"Dtype charts: {charts.dtype}")
        #print(f"Charts: {charts.shape}")
        #print(f"Dtype comb: {comb.dtype}")
        #print(f"Comb: {comb.shape}")
        #print(f"Tgt mask: {tgt_mask.shape}")

        # tgt, memory, tgt_mask=None
        # charts[:, :-1, :]
        out = self.transformer_decoder(charts[:, :-1, :], comb, tgt_mask).to(device)
        #print(f"Out transformer decoder shape: {out.shape}")
        out = self.linear(out).to(device)
        #print(f"Output linear: {output}")
        #print(f"Out linear shape: {out.shape}")
        return out


#token_sequenz
#indexer(token_sequenz) -> sequenz indices
#embedding_matrix[indexer(token_sequenz)]

# http://nlp.seas.harvard.edu/annotated-transformer/
class PositionalEncoding(nn.Module):
    "Implement the PE function."

    def __init__(self, d_model, dropout, max_len=5000):
        super(PositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        # Compute the positional encodings once in log space.
        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, d_model, 2) * -(math.log(10000.0) / d_model)
        )
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0)
        self.register_buffer("pe", pe)

    def forward(self, x):
        x = x + self.pe[:, : x.size(1)].requires_grad_(False)
        return self.dropout(x)



class Danceformer(nn.Module):
    def __init__(self, indexed_vocabulary):
        super().__init__()

        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        # Initialize cnn
        self.preprocessing_conv_input = Music_Net().to(device)
        # batch_size = 2
        # torch.Size([2, 3, 80, 500])
        #model_graph = draw_graph(self.preprocessing_conv_input, input_size=(batch_size, 3, 80, 500), device='meta')
        #model_graph.visual_graph
        #model_graph.render(directory=r'C:\Users\cassi\OneDrive\Desktop\Master_Thesis\music_net.png')


        # Initialize transformer,  https://pytorch.org/tutorials/beginner/transformer_tutorial.html
        # ntokens:20*3360 Initial src shape: torch.Size([3, 20, 3360])
        self.ntokens = len(indexed_vocabulary) # +1 for unidentified ...?
        self.indexed_vocabulary = indexed_vocabulary
        # len(indexed_vocabulary)  # len(preprocess_difficulty) = size of vocabulary, look into mlp_difficulty.py last layer output? cnn ouput?
        #self.ntokens_tgt = 320 # len chart....# len(indexed_vocabulary)  # len(preprocess_difficulty) = size of vocabulary, look into mlp_difficulty.py last layer output? cnn ouput?
        emsize = 60 #512  # 480?? dmodel??? zb 151 len(indexed_vocabulary)
        #emsize_tgt = 320 #512 # dmodel??? zb 151 len(indexed_vocabulary)
        d_hid = 200  # dimension of the feedforward network model in ``nn.TransformerEncoder``
        nlayers = 2  # number of ``nn.TransformerEncoderLayer`` in ``nn.TransformerEncoder``
        nhead = 2  # number of heads in ``nn.MultiheadAttention``
        dropout = 0.2  # dropout probability
        self.trans_model = TransformerModel(self.ntokens, emsize, nhead, d_hid, nlayers, dropout).to(device)
        # trans_model(data_imgs, diffs, charts) ([2, 20, 480]), ([2, 20]), ([2, 320])
        #trans_model_graph = draw_graph(self.trans_model, input_size=((2, 20, 480), (2, 20), (2, 320)), device='meta')
        #trans_model_graph.visual_graph
        #self.Difficulty_Net = Difficulty_Net()

        # Set parameters for training, from https://pytorch.org/tutorials/beginner/transformer_tutorial.html
        self.criterion = nn.CrossEntropyLoss().to(device)
        # lr = 5e-4  # learning rate
        # optimizer = torch.optim.SGD(trans_model.parameters(), lr=lr) #
        self.optimizer = torch.optim.AdamW(self.trans_model.parameters(), lr=0.0005)
        self.scheduler = torch.optim.lr_scheduler.StepLR(self.optimizer, 1.0, gamma=0.95)
        self.bptt = 35

        self.difficulty_net = Difficulty_Net()#.to(device)
        print("----------------------------------------")
        print("Initialized")
        print("----------------------------------------")



    def forward(self, imgs, diffs):
        # for each song in batch train a simple CNN and MLP to get preprocessed image and difficulty
        # imgs: (batch, channels, bands, time_frames)
        # example (2, 3, ...) stacked
        pre_conv_imgs = self.preprocessing_conv_input.forward(imgs)

        # Make MLP, preprocess difficulty to predict tensor
        preprocess_difficulty = self.difficulty_net.forward(diffs)

        return pre_conv_imgs, preprocess_difficulty
