# from https://deeplearning.neuromatch.io/projects/ComputerVision/spectrogram_analysis.html
# https://pytorch.org/tutorials/beginner/transformer_tutorial.html
import os
import glob
import random, shutil
import time
from io import StringIO

import simfile
import torch
import torch.nn as nn
import torch.nn.functional as F
# import torchvision.datasets as datasets
import torchvision.transforms as transforms
import numpy as np
import matplotlib.pyplot as plt
import IPython.display as display
import librosa
import librosa.display
import PIL.Image
from models.transformer import *
from SongPacksDataset import *
from tqdm.notebook import tqdm
from sklearn.neural_network import MLPRegressor
from preliminary_experiment import *
# from chart_embedding import *
from torch import nn, Tensor
from torch.nn import TransformerEncoder, TransformerEncoderLayer
import re

# other approach - not used here
def get_specs_diff_charts(song_dirs, indexed_vocabulary):
    # get data for each song: 5 tuple (sm file to charts and difficulties and 3 spectograms as tensors) per song
    all_specs = []
    all_charts = []
    all_difficulties = []
    max_len_spec = 0
    max_len_charts = 0

    for song_dir in song_dirs:
        specs = []
        # add sm file
        for item in os.listdir(song_dir):
            if item.endswith('.sm'):
                # specs.append(f"{song_dir}/{item}")
                with open(f"{song_dir}/{item}", 'r', encoding="utf8") as infile:
                    sm_file = simfile.load(infile)

                # get chart of middle difficulty
                # chart = get_middle_chart(sm_file.charts)
                charts = sm_file.charts
                # get difficulty
                difficulties = []
                # sample notes over time, return list of NoteData and [] if empy (for now)
                converted_charts = []
                for chart in charts:
                    notes_per_time = process_chart(sm_file, chart, indexed_vocabulary)
                    difficulty = chart.meter
                    difficulties.append(difficulty)
                    if max_len_charts < len(notes_per_time):
                        max_len_charts = len(notes_per_time)
                    converted_charts.append(notes_per_time)
                all_difficulties.append(difficulty)
                all_charts.append(converted_charts)
                # break
        for i in [23, 46, 93]:
            new_spec = torch.load(f"{song_dir}/mel_spectrogram_{i}.pt")
            new_spec = torch.squeeze(new_spec, dim=0)

            # sometimes shape of 3 windows not exactly the same ... to fix, only padding...
            # need to check other way around, case if less entries in previous entry
            if specs != [] and specs[-1].shape[1] != new_spec.shape[1]:
                missing_elements = specs[-1].shape[1] - new_spec.shape[1]
                padding = np.zeros((specs[-1].shape[0], missing_elements))
                new_spec = torch.from_numpy(np.concatenate((new_spec, padding), axis=1))

            # Determine maximum length
            spec_len = new_spec.shape[1]
            if max_len_spec < spec_len:
                max_len_spec = spec_len

            if specs != []:
                new_spec = new_spec.unsqueeze(0)
                specs = torch.cat((new_spec, specs), dim=0)
            else:
                specs.append(new_spec)
                specs = torch.stack(specs, dim=0)

        all_specs.append(specs)

    print(f"\n All_specs: 3 spectrograms as tensors, first entry shape: \n {all_specs[0].shape}")
    print(f"Dims all_specs: number songs = {len(all_specs)}, number items per song = {len(all_specs[0])}")
    # print(f"Difficulties per song: number songs = {len(difficulties)}")
    # print(f"Difficulties per song first entry = {difficulties[0]}")
    # print(f"Charts per song: number songs = {len(charts)}")
    # print(f"Charts per song first entry = {charts[0]}")
    # 512 or 1024 tokens standard
    print(f"Maximum length spectogram = {max_len_spec}")
    print("Usually standard maximum length token sequence for transformer 512 or 1024 tokens")
    print(f"Maximum length chart = {max_len_charts}")
    return all_specs, all_difficulties, all_charts, max_len_spec, max_len_charts

# not used
def unify_specs(all_specs, max_len_spec):
    # pad all tensors to have same length
    all_specs_unified = []
    for song in all_specs:
        song_data = []
        for spec in song:
            new_song_data = [torch.nn.functional.pad(x, pad=(0, max_len_spec - x.numel()), mode='constant', value=0)
                             for x in spec]
            song_data.append(new_song_data)
            # print(len(new_song_data)) # stereo to mono ...
        all_specs_unified.append(song_data)
        # does not work the way it should, look
        # format mel_specs given... look there how to unify

    print(f"All_specs_unified: 3 spectrograms as tensors, first entry: \n {all_specs_unified[0]}")
    print(
        f"Dims all_specs_unified: number songs = {len(all_specs_unified)}, number items per song = {len(all_specs_unified[0])}")

    return all_specs_unified


# not used
def unify_charts(charts, max_len_charts):
    charts_unified = []
    for chart in charts:
        chart_data = chart
        if max_len_charts - len(chart) > 0:
            # as long as possible (divideable by 4) append ,0000
            max_iters_default = int(max_len_charts - len(chart) / 4)
            for i_df in range(max_iters_default):
                # chart_data = f'{chart}\n,\n0000'
                chart_data = f'{chart}\n0000'
            # in the end rest possible number (1,2,3) of zeros, relevant??
            # if max_len_charts - len(chart) % 4 == 0:
            #    chart_data = f'{chart}'.ljust(max_len_charts - len(chart), ',\n0000')
            if (max_len_charts - len(chart)) % 4 == 1:
                chart_data = f'{chart}\n000'
            if (max_len_charts - len(chart)) % 4 == 2:
                chart_data = f'{chart}\n00'
            if (max_len_charts - len(chart)) % 4 == 3:
                chart_data = f'{chart}\n0'
            # print("modified chart:")
            # print(chart_data)
            # print((max_len_charts - len(chart)) % 4)
        charts_unified.append(chart_data)

    print(f"Charts per song: number songs = {len(charts_unified)}")
    # print(f"Charts per song first entry = {charts[0]}")
    return charts_unified
