# from https://deeplearning.neuromatch.io/projects/ComputerVision/spectrogram_analysis.html
# https://pytorch.org/tutorials/beginner/transformer_tutorial.html
import os
import glob
import random, shutil
import time
from datetime import datetime, timedelta
from io import StringIO
import simfile
import soundfile
import torch
import torch.nn as nn
import torch.nn.functional as F
# import torchvision.datasets as datasets
import torchvision.transforms as transforms
import numpy as np
import matplotlib.pyplot as plt
import librosa
import librosa.display
from models.transformer import *
from SongPacksDataset import *
from preliminary_experiment import *
from simfile.timing.engine import TimingEngine
from torch import nn, Tensor
from torch.nn import TransformerEncoder, TransformerEncoderLayer
import re
import torchaudio.transforms as T


def filter_whitespaces(notes):
    # regular expression pattern matching multiple spaces between \n
    pattern = re.compile(r'\n\s*\n')

    filtered_notes = pattern.sub('\n', notes)

    return filtered_notes


def index_tokens(token_sequence):
    # Create an index for the tokens
    token_index = {token: index for index, token in enumerate(token_sequence)}

    return token_index


def built_chart_vocabulary(song_dirs):
    vocabulary_set = set()
    cols_init = 4
    # once add zero string
    note_string = ["0"] * cols_init
    notedata_new = StringIO()
    notedata_new.write("".join(note_string))
    # print(notedata_new.getvalue())
    vocabulary_set.update([notedata_new.getvalue()])
    # special token/unknown token
    for num, song_dir in enumerate(song_dirs):
        if num % 1000 == 0:
            print(f"Song: {num}")
        for item in os.listdir(song_dir):
            if item.endswith('.sm'):
                # specs.append(f"{song_dir}/{item}")
                with open(f"{song_dir}/{item}", 'r', encoding="ISO-8859-1") as infile:
                    sm_file = simfile.load(infile)

                # get chart of middle difficulty
                # chart = get_middle_chart(sm_file.charts)
                charts = sm_file.charts
                # print("Built vocabulary")
                for chart in charts:
                    # notes is string
                    note_data = NoteData(chart)
                    timing_data = TimingData(sm_file)
                    cols = note_data.columns

                    # dictionary
                    tmp = dict()
                    for timed_note in time_notes(note_data, timing_data):
                        if timed_note.time in tmp.keys():
                            tmp[timed_note.time] = tmp[timed_note.time][:timed_note.note.column] + str(
                                timed_note.note) + \
                                                   tmp[timed_note.time][timed_note.note.column + 1:]
                        else:
                            note_string = ["0"] * cols
                            note_string[timed_note.note.column] = str(timed_note.note)
                            note_string_new = StringIO()
                            note_string_new.write("".join(note_string))
                            tmp[timed_note.time] = note_string_new.getvalue()

                    vocabulary_set.update(tmp.values())

    print("----------------------------------")
    print("Built Vocabulary:")
    print(vocabulary_set)
    indexed_vocabulary = index_tokens(vocabulary_set)
    print("Indexing:")
    print(indexed_vocabulary)
    print("Length vocabulary:")
    print(len(indexed_vocabulary))
    return indexed_vocabulary


def mixing_down(waveform):
    # if not mono
    if waveform.shape[0] > 1:
        waveform = torch.mean(waveform, dim=0, keepdim=True)
    return waveform


def get_paths_diff_charts(safe_dir, song_dirs, indexed_vocabulary, device):
    # get data for each song: 5 tuple (sm file to charts and difficulties and 3 spectograms as tensors) per song
    # want 1song, 1chart, 1diff
    # multiple entries possible for a song (per number charts/diffs)
    # could look for idx in folder, if files exist skip

    max_len_charts = 0
    num_songs = len(song_dirs)
    actual_num_songs = 0
    num_charts = 0
    print(f"Number of songs to process: {num_songs}")
    start_time_preprocess = time.time()
    files_with_value_error = []
    files_with_lib_error = []
    files_with_value_error_number = 0
    files_with_lib_error_number = 0
    num_files = 0
    # in ms
    win_length_list_ms = [23, 46, 93]
    #empty_token = indexed_vocabulary.get("0000")
    # alternativ vorletzter und letzter vocab eintrag, value
    #start_token = indexed_vocabulary.get("<s>")
    #end_token = indexed_vocabulary.get("<e>")


    for idx, song_dir in enumerate(song_dirs):
        # safe per song, then write to file, and new song
        all_waves = []
        all_sample_rates = []
        all_paths = []
        all_charts = []
        all_difficulties = []
        specs_23 = []
        specs_46 = []
        specs_93 = []

        # add sm file
        # should be set in following loop
        number_of_charts_diffs = 0
        all_paths_old_len = len(all_paths)
        audio_exists = False
        list_items = os.listdir(song_dir)
        # no runthrough, need at least 2 files, sm and audio file
        if len(list_items) >= 2:
            charts_per_song = 0
            for item in os.listdir(song_dir):
                if item.endswith('.sm'):
                    opened_file = simfile.open(f"{song_dir}/{item}")
                    # specs.append(f"{song_dir}/{item}")
                    with open(f"{song_dir}/{item}", 'r', encoding="ISO-8859-1") as infile:
                        sm_file = simfile.load(infile)

                    # get chart of middle difficulty
                    # chart = get_middle_chart(sm_file.charts)
                    charts = sm_file.charts
                    # get difficulty
                    difficulties = []
                    # sample notes over time, return list of NoteData and [] if empy (for now)
                    converted_charts = []
                    charts_per_song = len(charts)

                    # if idx > 2500:
                    #    print(
                    #        f"File: {item}, Number_charts: {len(charts)}, Song: {idx}, Processed charts: {num_charts}, "
                    #        f"Elapsed time: {elapsed_second_step}, Time: {now}")
                    for i, chart in enumerate(charts):
                        num_charts += 1
                        notes_per_time = process_chart(opened_file, sm_file, chart, indexed_vocabulary, device)
                        difficulty = chart.meter
                        difficulties.append(int(difficulty))
                        if max_len_charts < len(notes_per_time):
                            max_len_charts = len(notes_per_time)
                        converted_charts.append(notes_per_time)
                    # if idx > 2500:
                    #     print(
                    #         f"File: {item}, Chart: {i}, "
                    #         f"Elapsed time: {elapsed_second_step}, Time: {now}")
                    all_difficulties.extend(difficulties)
                    all_charts.extend(converted_charts)
                    assert (len(converted_charts) == len(difficulties))
                    number_of_charts_diffs = len(converted_charts)
                if item.endswith('.ogg') or item.endswith('.mp3'):
                    # be sure not two of one file in two formats
                    if not audio_exists:
                        melspec_old_23_len = specs_23
                        melspec_old_46_len = specs_46
                        melspec_old_93_len = specs_93
                        all_waves_old_len = len(all_waves)
                        all_sample_rates_old_len = len(all_sample_rates)
                        all_paths_old_len = len(all_paths)
                        actual_num_songs_old = actual_num_songs
                        try:
                            # here right version of torchaudio/torch needed ... different with cuda
                            waveform, sample_rate = torchaudio.load(f"{song_dir}/{item}")
                            # print(f"Sample rate: {sample_rate}")
                            # waveform = sampling_down(waveform, sample_rate, target_sample_rate)
                            waveform = mixing_down(waveform)

                            win_length_list = []
                            for length in win_length_list_ms:
                                win_length_list.append(int(sample_rate * (length / 1000)))

                            for j, win_length in enumerate(win_length_list):
                                hop_length = int(sample_rate * (10 / 1000))  # win length = 10, 10ms aka stride
                                n_mels = 80  # Number of mel filterbanks 80
                                n_fft = win_length

                                # https://pytorch.org/audio/stable/generated/torchaudio.functional.melscale_fbanks.html#torchaudio.functional.melscale_fbanks
                                # https://pytorch.org/audio/stable/generated/torchaudio.transforms.MelSpectrogram.html#torchaudio.transforms.MelSpectrogram
                                mel_spectrogram_transform = T.MelSpectrogram(
                                    sample_rate=sample_rate,
                                    n_fft=n_fft,
                                    win_length=win_length,
                                    hop_length=hop_length,
                                    center=True,
                                    pad_mode="reflect",
                                    power=2.0,
                                    norm="slaney",
                                    n_mels=n_mels,
                                    mel_scale="htk",
                                )

                                mel_spectrogram = mel_spectrogram_transform(waveform)
                                if win_length_list_ms[j] == 23:
                                    specs_23.append(mel_spectrogram)
                                elif win_length_list_ms[j] == 46:
                                    specs_46.append(mel_spectrogram)
                                else:
                                    specs_93.append(mel_spectrogram)
                                # torch.save(mel_spectrogram, f"{safe_dir}/melspecs_{win_length_list_ms[j]}/{idx}.pt")

                            # all_paths.append([waveform, sample_rate, f"{song_dir}/{item}"])
                            all_waves.append(waveform)
                            all_sample_rates.append(sample_rate)
                            all_paths.append(f"{song_dir}/{item}")
                            actual_num_songs += 1
                            audio_exists = True
                        except ValueError:
                            files_with_value_error.append(f"{song_dir}/{item}")
                            files_with_value_error_number += 1
                            # undo possible changes
                            # happens at most once
                            if len(specs_23) > melspec_old_23_len:
                                specs_23 = specs_23[:-1]
                            if len(specs_46) > melspec_old_46_len:
                                specs_46 = specs_46[:-1]
                            if len(specs_93) > melspec_old_93_len:
                                specs_93 = specs_93[:-1]
                                # happens up to 3 times..
                            if len(all_waves) > all_waves_old_len:
                                all_waves = all_waves[:all_waves_old_len]
                            if len(all_paths) > all_paths_old_len:
                                all_paths = all_paths[:all_paths_old_len]
                            if len(all_sample_rates) > all_sample_rates_old_len:
                                all_sample_rates = all_sample_rates[:all_sample_rates_old_len]
                            actual_num_songs = actual_num_songs_old
                            audio_exists = False
                        except soundfile.LibsndfileError:
                            files_with_lib_error.append(f"{song_dir}/{item}")
                            files_with_lib_error_number += 1
                            # undo possible changes
                            # happens at most once
                            if len(specs_23) > melspec_old_23_len:
                                specs_23 = specs_23[:-1]
                            if len(specs_46) > melspec_old_46_len:
                                specs_46 = specs_46[:-1]
                            if len(specs_93) > melspec_old_93_len:
                                specs_93 = specs_93[:-1]
                                # happens up to 3 times..
                            if len(all_waves) > all_waves_old_len:
                                all_waves = all_waves[:all_waves_old_len]
                            if len(all_paths) > all_paths_old_len:
                                all_paths = all_paths[:all_paths_old_len]
                            if len(all_sample_rates) > all_sample_rates_old_len:
                                all_sample_rates = all_sample_rates[:all_sample_rates_old_len]
                            actual_num_songs = actual_num_songs_old
                            audio_exists = False
                        except MemoryError:
                            print("Memory Error")
                            print(f"{song_dir}/{item}")
                            # undo possible changes
                            #happens at most once
                            if len(specs_23) > melspec_old_23_len:
                                specs_23 = specs_23[:-1]
                            if len(specs_46) > melspec_old_46_len:
                                specs_46 = specs_46[:-1]
                            if len(specs_93) > melspec_old_93_len:
                                specs_93 = specs_93[:-1]
                                # happens up to 3 times..
                            if len(all_waves) > all_waves_old_len:
                                all_waves = all_waves[:all_waves_old_len]
                            if len(all_paths) > all_paths_old_len:
                                all_paths = all_paths[:all_paths_old_len]
                            if len(all_sample_rates) > all_sample_rates_old_len:
                                all_sample_rates = all_sample_rates[:all_sample_rates_old_len]
                            actual_num_songs = actual_num_songs_old
                            audio_exists = False
                            # exit(1)

            # be sure not two music files added
            assert (len(all_paths) <= all_paths_old_len + 1)

            # something changed, but not one single music file/ not audio_exists
            if number_of_charts_diffs > 0 and len(all_paths) == all_paths_old_len:
                # remove changes, forget this song
                # for number in range(number_of_charts_diffs):
                all_difficulties = all_difficulties[:-number_of_charts_diffs]
                all_charts = all_charts[:-number_of_charts_diffs]
                num_charts = num_charts - charts_per_song

            # if something changed and one music file added, add same path(last entry) for number of changes
            if number_of_charts_diffs > 0 and len(all_paths) == all_paths_old_len + 1:
                for number in range(number_of_charts_diffs - 1):
                    all_paths.append(all_paths[-1])
                    all_waves.append(all_waves[-1])
                    specs_23.append(specs_23[-1])
                    specs_46.append(specs_46[-1])
                    specs_93.append(specs_93[-1])
                    all_sample_rates.append(all_sample_rates[-1])

            # if no chart found but audio, remove audio, can be at most one more audio/path according check before
            # -> remove last filepath
            if number_of_charts_diffs == 0 and len(all_paths) > all_paths_old_len:
                all_paths = all_paths[:-1]
                all_waves = all_waves[:-1]
                specs_23 = specs_23[:-1]
                specs_46 = specs_46[:-1]
                specs_93 = specs_93[:-1]
                all_sample_rates = all_sample_rates[:-1]
                actual_num_songs -= 1

            # verify number items is equal
            # print(
            #    f"Number paths: {len(all_paths)}, Number diffs: {len(all_difficulties)}, Number charts: {len(all_charts)}")
            try:
                assert len(all_paths) == len(all_difficulties) == len(all_charts)
                assert len(all_paths) == len(all_waves) == len(all_sample_rates)
                assert len(all_waves) == len(specs_23) == len(specs_46) == len(specs_93)
            except AssertionError:
                print(
                    f'Len paths {len(all_paths)}, Len difficulties {len(all_difficulties)}, Len charts {len(all_charts)}')
                print(f'Len waves {len(all_waves)}, Len sample rates {len(all_sample_rates)}, Len specs_23 {len(specs_23)}, '
                      f'Len specs_46 {len(specs_46)}, Len specs_93 {len(specs_93)}')
                print(f'Len paths old {all_paths_old_len},  number_of_charts_diffs {number_of_charts_diffs}')
                print(f'List items {list_items}')
                exit(1)

            if idx % 100 == 0:
                elapsed_second_step = time.time() - start_time_preprocess
                now = datetime.now()
                # index starts at 0, plus 1 for number of song processed, actual song number is for finished processed up to then
                print(
                    f"Song: {idx + 1}, Actual song: {actual_num_songs}, Processed charts: {num_charts}, "
                    f"Elapsed time: {str(timedelta(seconds=elapsed_second_step))}, Time: {now}, "
                    f"Number error files: {files_with_value_error_number + files_with_lib_error_number}")

            if all_waves != [] and all_difficulties != [] and all_charts != [] and specs_23 != [] and specs_46 != [] and specs_93 != []:
                # safe each datapoint to file
                # all_waves = torch.stack(all_waves, dim=0)
                # all_waves = torch.squeeze(all_waves, 1)
                # all_waves = torch.stack(all_waves)
                # all_sample_rates = torch.tensor(all_sample_rates)
                # all_difficulties = torch.tensor(all_difficulties)
                # all_charts = torch.stack([torch.tensor(cha) for cha in all_charts])

                # save each datapoint
                for num in range(len(all_charts)):
                    with open(f'{safe_dir}/all_paths/{num_files}.txt', 'w') as f:
                        f.write(f"{all_paths[num]}")

                    with open(f'{safe_dir}/all_waves/{num_files}.pkl', 'wb') as fp:
                        # with open(r'/scratch/grzonkow/vocabulary_model.pkl', 'wb') as fp:
                        pickle.dump(all_waves[num], fp)

                    torch.save(all_sample_rates[num], f'{safe_dir}/all_sample_rates/{num_files}.pt')
                    torch.save(all_difficulties[num], f'{safe_dir}/all_difficulties/{num_files}.pt')
                    torch.save(specs_23[num], f'{safe_dir}/mel_specs_23/{num_files}.pt')
                    torch.save(specs_46[num], f'{safe_dir}/mel_specs_46/{num_files}.pt')
                    torch.save(specs_93[num], f'{safe_dir}/mel_specs_93/{num_files}.pt')

                    with open(f'{safe_dir}/all_charts/{num_files}.pkl', 'wb') as fp:
                        # with open(r'/scratch/grzonkow/vocabulary_model.pkl', 'wb') as fp:
                        pickle.dump(all_charts[num], fp)

                    num_files += 1

    # print(f"\n Number all paths: number songs = {len(all_paths)}, first 5 paths = {all_paths[:5]}")
    # print(f"Difficulties per song: number songs = {len(difficulties)}")
    # print(f"Difficulties per song first entry = {difficulties[0]}")
    # print(f"Charts per song: number songs = {len(all_charts)}")
    # print(f"Charts per song first entry = {all_charts[0]}")
    # 512 or 1024 tokens standard
    # print("Usually standard maximum length token sequence for transformer 512 or 1024 tokens")
    # print(f"Maximum length chart = {max_len_charts}")
    print(
        f"Song: {idx + 1}, Actual song: {actual_num_songs}, Processed charts: {num_charts}, "
        f"Elapsed time: {str(timedelta(seconds=elapsed_second_step))}, Time: {now}, "
        f"Number error files: {files_with_value_error_number + files_with_lib_error_number}, "
        f"Number of value error: {files_with_value_error_number}, "
        f"Number with lib error: {files_with_lib_error_number}, "
        f"Value error files: {files_with_value_error}, "
        f"Lib error files: {files_with_lib_error}")

    # torch.load()

    return  # all_paths, all_difficulties, all_charts, max_len_charts


# Preprocess sm chart into uniform distant time chart of notes
def process_chart(openedfile, test_file, chart, indexed_vocabulary, device):
    # time notes of chart and preserve info about note
    # sampling right amount of chart notes for example 1/8 (0.125)second, starting point...
    timing_data = TimingData(test_file)
    note_data = NoteData(chart)
    cols = note_data.columns

    # empty_token = indexed_vocabulary.get("0000")
    # alternativ vorletzter und letzter vocab eintrag, value
    start_token = indexed_vocabulary.get("<s>")
    end_token = indexed_vocabulary.get("<e>")

    # timing_data_test = TimingData(openedfile)
    # Converting song time to beats
    # engine = TimingEngine(timing_data_test)

    # print("Timed notes: ")
    # count = 0
    # lenght of song
    # every 0.125 sec save note, if non existent 0? start note?
    # format: column (
    # first_note = None
    # timed_notes = []

    tmp = dict()
    # last_note_time = 0

    for timed_note in time_notes(note_data, timing_data):
        # if last_note_time < timed_note.time:
        #    last_note_time = timed_note.time
        if timed_note.time in tmp.keys():
            tmp[timed_note.time] = tmp[timed_note.time][:timed_note.note.column] + str(timed_note.note) + \
                                   tmp[timed_note.time][timed_note.note.column + 1:]
        else:
            note_string = ["0"] * cols
            note_string[timed_note.note.column] = str(timed_note.note)
            note_string_new = StringIO()
            note_string_new.write("".join(note_string))
            tmp[timed_note.time] = note_string_new.getvalue()

    tmp = dict(sorted(tmp.items()))

    # for timed_note in time_notes(note_data, timing_data):
    # count += 1
    # if first_note is None:
    #    first_note = timed_note
    # if timed_note.note.note_type == NoteType.TAIL:
    # print(timed_note)
    # timed_notes.append(timed_note)
    # if timed_note.time in timed_notes.keys():
    #
    # else:
    #    tmp[timed_note.time] =

    # print(f"Number timed notes: {count}")
    # print(f"First timed notes: {first_note}")

    # generally, need fixed number of intervalls from 0
    # notes_per_time = []
    range_dist = 1 / 64  # 1/8 too many notes in interval
    # first_interval = first_note.time / range_dist
    # print(f"First interval: {first_interval}")
    # start from first note or 0? until song end not last timed note? first_note.time -> need time_offset
    # for time_step in np.arange(0, timed_notes[-1].time, range_dist):

    # start with zero values in intervals
    # since no update of empty string, map no matter how many columns to same empty note with 4 columns
    # note_string = ["0"] * cols
    note_string = ["0"] * 4
    notedata = StringIO()
    notedata.write("".join(note_string))
    empty_note = notedata.getvalue()
    empty_note_encoded = indexed_vocabulary[empty_note]

    # notes_in_range = [note_string] * len(np.arange(0, tmp[-1].key(), range_dist))
    # notes_in_range = dict()
    notes_in_range = []
    notes_in_range.append(start_token)

    # print(list(tmp.keys())[-1])
    # count_more_notes = 0

    for time_step in np.arange(0, list(tmp.keys())[-1], range_dist):
        # notes_per_range = []
        # notes_per_range_info = []
        # counter = 0
        # for key in timed_notes:
        # only at one time, jumps?
        res = [v for k, v in tmp.items() if time_step > k >= time_step - range_dist]
        if not res:
            # notes_in_range[time_step] = empty_note_encoded
            notes_in_range.append(empty_note_encoded)
        else:
            # what to do if more then one note in interval? combine could yield new notes not in dict ...
            # take first one
            # if len(res) > 1:
            #    count_more_notes += 1
            # notes_in_range[time_step] = indexed_vocabulary[res[0]]
            notes_in_range.append(indexed_vocabulary[res[0]])

    # take care of append 0 case in loading data random
    notes_in_range.append(end_token)

    # print(f"Number counted more notes in interval: {count_more_notes}")
    # if time_step > tmp.keys() >= time_step - range_dist:
    #         notes_per_range.append(timed_note.note)
    #         notes_per_range_info.append(timed_note)
    # counter += 1
    # if counter != 0:
    # print(f"Number of timed notes in interval: {counter}")
    # if len(notes_per_range)>=2:
    #    print(f"Notes per range: {notes_per_range}")
    #    print(f"Notes per range: {notes_per_range_info}")

    # if notes_per_range != []:
    #   notes_per_range = str(NoteData.from_notes(notes_per_range, cols))
    # looking at simfile programming from_notes

    # if list not empty, if more than one lets see... to fix?
    # problem: new notes if combinations over ranges, ..., precompute all in this way in voc,
    # or other voc building, all combis, fix!
    # easy variant: just take first one

    # empty list, keep empty token
    # if notes_per_range_info != []:
    #     # one element list, transform to 1 token
    #     if len(notes_per_range_info) == 1:
    #         note_string[notes_per_range_info[0].note.column] = str(notes_per_range_info[0].note)
    #     # more than 1 note, at least 2, compare if equal time and beat
    #     else:
    #         #quick_fix = False
    #
    #         for i, note in enumerate(notes_per_range_info):
    #             for j in range(i+1, len(notes_per_range_info)):
    #                 note_2 = notes_per_range_info[j]
    #                 if note.time == note_2.time: #and note.note.beat == note_2.note.beat:
    #                     note_string[note.note.column] = str(note.note)
    #                     note_string[note_2.note.column] = str(note_2.note)

    # if note_string == ['M', 'M', '0', '0']:
    #    print(note_string)
    # if beatchange .... what?, do not combine... leave one out?? quick take first one, print
    # ignore for now, check later again!
    # if note.time == note_2.time and note.note.beat != note_2.note.beat:
    #    print(f"Problem note basis: {notes_per_range_info}")
    #    print(f"Problem note hittable: {engine.hittable(note.note.beat)}")
    #    print(f"Problem note 2 hittable: {engine.hittable(note_2.note.beat)}")
    # note_string[note.note.column] = str(note.note)
    # print(f"Problem note note quick fix: {note_string}")
    #    quick_fix = True
    # if quick_fix == True and no

    # if str(note) == "1" or str(note) == "2":
    #    note_string[note.column] = str(note)

    # if not empty, take first note, easy variant
    # if notes_per_range:
    #    notes_per_range = notes_per_range.pop()

    # notedata = StringIO()
    # notedata.write("".join(note_string))
    # notedata.getvalue()
    # notedata.write("\n")
    # note_strings = ["0"] * columns
    #             for note in row:
    #                 note_strings[note.column] = str(note)
    #             notedata.write("".join(note_strings))
    #             notedata.write("\n")
    # somehow NoteData format of empty sequence such as 0000, ...
    # else:
    # or just
    #    notes_per_range = "0000"
    # note_data_representation = np.zeros(cols)
    # note_data_representation = Note(beat=Beat(0, 2), column=i%cols, note_type=NoteType.)
    #    note_data_representation = []
    # notes_per_time.append(notedata.getvalue())
    # print(f"Notes per time first 20 entries: {notes_per_time[:20]}")
    # print(f"Notes per time last 20 entries: {notes_per_time[:-20]}")
    # print(f"Length of notes per time: {len(notes_per_time)}")

    # convert list of lists to list of numbers -> tensor
    # as original notes did? 4 integers e.g. 0000, 0010, ...
    # position/column: left, right, up, down
    # number: 1 (tab), 2, 3 (hold...), 4, ..
    # print(notes_in_range.values())
    # converted_notes_per_time = [indexed_vocabulary[str(token)] for token in notes_in_range.values()]
    # print(f"converted_notes_per_time first 20 entries: {converted_notes_per_time[:20]}")
    # print(f"converted_notes_per_time first 64 entries starting later intervall -1 : "
    #      f"{converted_notes_per_time[int(first_interval)-1:int(first_interval)+64]}")
    # print(f"notes_per_time first 64 entries starting later intervall -1 : "
    #      f"{notes_per_time[int(first_interval) - 1:int(first_interval) + 64]}")
    # print(f"converted_notes_per_time last 20 entries: {converted_notes_per_time[:-20]}")
    # print(f"Length of converted_notes_per_time: {len(converted_notes_per_time)}")

    # print("---------------------------------------------------------------------------------")
    return notes_in_range  # notes_per_time
