import argparse
import matplotlib.pyplot as plt
import torch
import numpy as np


def setup_parser():
    out = argparse.ArgumentParser()
    out.add_argument('--folder',
                     default=r'C:/Users/cassi/OneDrive/Desktop/Master_Thesis/2702_test/',
                     type=str, help="Path to load model parameter")
    return out


def plot(data, label, save_path):
    fig = plt.figure()
    data = np.array(data)
    plt.plot(np.arange(0, len(data)), data, label=label)
    plt.legend()
    plt.savefig(save_path)


if __name__ == '__main__':
    parser = setup_parser()
    args, unknown = parser.parse_known_args()
    device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')

    folder = args.folder


    with open(f"{folder}avg_prob_non_empty_per_batch.txt", 'r') as fp:
        avg_prob_non_empty_pre = fp.read()

    avg_prob_non_empty_pre = avg_prob_non_empty_pre.split("\n")
    avg_prob_non_empty = [float(t) for t in avg_prob_non_empty_pre[:-1]]

    plot(avg_prob_non_empty, "avg_prob_non_empty", f"{folder}avg_prob_non_empty.png")

    with open(f"{folder}ms_per_batch.txt", 'r') as fp:
        ms_per_batch_all_pre = fp.read()

    ms_per_batch_all_pre = ms_per_batch_all_pre.split("\n")
    ms_per_batch_all = [float(t) for t in ms_per_batch_all_pre[:-1]]

    plot(ms_per_batch_all, "ms_per_batch_all", f"{folder}ms_per_batch_all.png")

    with open(f"{folder}ppl.txt", 'r') as fp:
        ppl_all_pre = fp.read()

    ppl_all_pre = ppl_all_pre.split("\n")
    ppl_all = [float(t) for t in ppl_all_pre[:-1]]

    plot(ppl_all[1:], "ppl_all", f"{folder}ppl_all.png")

    with open(f"{folder}ppl_v_2.txt", 'r') as fp:
        ppl_v_2_all_pre = fp.read()

    ppl_v_2_all_pre = ppl_v_2_all_pre.split("\n")
    ppl_v_2_all = [float(t) for t in ppl_v_2_all_pre[:-1]]

    plot(ppl_v_2_all[7:], "ppl_v_2_all", f"{folder}ppl_v_2_all.png")

    with open(f"{folder}loss.txt", 'r') as fp:
        loss_all_pre = fp.read()

    loss_all_pre = loss_all_pre.split("\n")
    loss_all = [float(t) for t in loss_all_pre[:-1]]

    plot(loss_all, "loss_all", f"{folder}loss_all.png")

    try:
        with open(f"{folder}greater_ratio_values.txt", 'r') as fp:
            greater_ratio_values_all_pre = fp.read()

        greater_ratio_values_all_pre = greater_ratio_values_all_pre.split("\n")
        greater_ratio_values_all = [float(t) for t in greater_ratio_values_all_pre[:-1]]

        plot(greater_ratio_values_all, "greater_ratio_values_all", f"{folder}greater_ratio_values_all.png")
    except FileNotFoundError:
        greater_ratio_values_all = []

    with open(f"{folder}avg_prob_non_empty_per_batch.txt", 'r') as fp:
        avg_prob_non_empty_per_batch_pre = fp.read()

    avg_prob_non_empty_per_batch_pre = avg_prob_non_empty_per_batch_pre.split("\n")
    avg_prob_non_empty_per_batch = [float(t) for t in avg_prob_non_empty_per_batch_pre[:-1]]

    plot(avg_prob_non_empty_per_batch, "avg_prob_non_empty_per_batch", f"{folder}avg_prob_non_empty_per_batch.png")

    with open(f"{folder}average_right_tokens.txt", 'r') as fp:
        average_right_tokens_all_pre = fp.read()

    average_right_tokens_all_pre = average_right_tokens_all_pre.split("\n")
    average_right_tokens_all = [float(t) for t in average_right_tokens_all_pre[:-1]]

    plot(average_right_tokens_all, "average_right_tokens_all", f"{folder}average_right_tokens_all.png")

