import math
import os
import pickle

import simfile
import torchaudio.transforms as T
import torch
from simfile.timing import TimingData

if __name__ == '__main__':
    safe_dir = "/scratch/grzonkow/output"
    #safe_dir = "C:/Users/cassi/OneDrive/Desktop/Master_Thesis/output"
    num_files = len(os.listdir(f"{safe_dir}/all_paths/"))
    # in ms
    win_length_list_ms = [23, 46, 93]
    for win in win_length_list_ms:
        os.makedirs(f"{safe_dir}/mel_specs_{win}/", exist_ok=True)

    for idx in range(num_files):
        with open(f'{safe_dir}/all_waves/{idx}.pkl', 'rb') as fp:
            # with open(r'/scratch/grzonkow/vocabulary_model.pkl', 'wb') as fp:
            waveform = pickle.load(fp)

        sample_rate = torch.load(f"{safe_dir}/all_sample_rates/{idx}.pt")

        with open(f"{safe_dir}/all_paths/{idx}.txt", 'r', encoding='ISO-8859-1') as f:
            path = f.read()

        # look at custom_collate_fn
        win_length_list = []
        for length in win_length_list_ms:
            win_length_list.append(int(sample_rate * (length / 1000)))

        length_song = waveform.shape[1] / sample_rate
        #print(f"Start: {start}")

        for j, win_length in enumerate(win_length_list):
            # win_length = int(sample_rate * (23 / 1000))  # 23ms, 46ms, and 93m and int(sample_rate * (length / 1000))
            hop_length = int(sample_rate * (10 / 1000))  # 10ms aka stride
            n_mels = 80  # Number of mel filterbanks 80
            n_fft = win_length

            # https://pytorch.org/audio/stable/generated/torchaudio.functional.melscale_fbanks.html#torchaudio.functional.melscale_fbanks
            # https://pytorch.org/audio/stable/generated/torchaudio.transforms.MelSpectrogram.html#torchaudio.transforms.MelSpectrogram
            mel_spectrogram_transform = T.MelSpectrogram(
                sample_rate=sample_rate,
                n_fft=n_fft,
                win_length=win_length,
                hop_length=hop_length,
                center=True,
                pad_mode="reflect",
                power=2.0,
                norm="slaney",
                n_mels=n_mels,
                mel_scale="htk",
            )

            #print(path[path.rindex('.')+1:])
            #print(path[:path.rindex('.')+1])
            #print(path[:path.rindex('.')+1] + 'sm')
            #opened_file = simfile.open(path[:path.rindex('.')+1] + 'sm')
            #with open(f"{song_dir}/{item}", 'r', encoding="ISO-8859-1") as infile:
            #    sm_file = simfile.load(infile)
            #timing_data = TimingData(opened_file)
            # (number_samples/second) / (int((number_samples/second) * (10 / 1000))) # 10ms aka stride
            #base_frames = sample_rate / hop_length
            #print(f"base frame: {base_frames}")
            # hop length of 10ms in frames
            #print(f"hop length: {hop_length}")
            #frame = time * base_frames # (in melspectogram) frame per second

            mel_spectrogram = mel_spectrogram_transform(waveform)
            #print(f"last base frame: {(mel_spectrogram.shape[2]-1)/base_frames}")
            #print(f"ratio base frame: {base_frames/(mel_spectrogram.shape[2]-1)}")
            #print(mel_spectrogram.shape)

            # y axis frequenccy bands = 80
            # x axis frames in melspec
            # -> transform x axis to time with time = frame/base_frame
            #time_sequence = []
            #for i in range(mel_spectrogram.shape[2]):
            #    time_sequence.append(i/base_frames)

            #time_sequence = torch.tensor(time_sequence)
            #print(time_sequence.shape)
            # checked for song 110, starts at 0, goes till the end of 126,54seconds=2,6min
            # 0.010 seconds steps aka 10ms stride/hop_length -> 100 steps for 1 second
            #print(time_sequence[:10])
            #print(time_sequence[-10:])
            # Perform transform
            # Mel frequency spectrogram of size (…, n_mels, time) first channels?
            # if item[0].shape[2] <= max_len_specs:
            #    specs_unified = torch.stack([pad(item[0], max_len_specs)])
            # else:
            # print(f"Sliced waveform shape: {waveform1.shape}")
            #mel_spectrogram = mel_spectrogram_transform(waveform)
            # initializing separator, for saving slice in right folder
            # print(path)
            torch.save(mel_spectrogram, f"{safe_dir}/melspecs_{win_length_list_ms[j]}/{idx}.pt")
            # torch.save(mel_spectrogram, f"{result}/mel_spectrogram_{win_length_list_ms[j]}.pt")
            # if mel_spectrogram.size()[2] > max_frame_size:
            #    max_frame_size = mel_spectrogram.size()[2]

            # # test cut
            # max_len_specs = 5
            # # in seconds
            # start = torch.randint(int(length_song - max_len_specs), [1])
            # # print(f"Start: {start}")
            # # Slicing, start, max_len_specs in seconds
            # # repeatation, read sample rate out of file
            # #hop_length = (int(sample_rate * (10 / 1000)))  # 10ms aka stride
            # sek_hop_length = (10 / 1000)
            # #base_frames = sample_rate / hop_length
            #
            # s_ind = int(start/sek_hop_length)
            # e_ind = int((start+max_len_specs)/sek_hop_length)
            # print(f"Start: {start}")
            # print(f"Start index: {s_ind}")
            # print(f"End: {start+max_len_specs}")
            # print(f"End index: {e_ind}")
            # print(f"Time sequence at that time: {time_sequence[s_ind:e_ind]}")
            # sub_mel_spectogram = mel_spectrogram[:, :, s_ind:e_ind]
            # print(sub_mel_spectogram.shape)
            # #frame_offset, num_frames = start * sample_rate, sample_rate * max_len_specs
            # #sub_mel_spectogram = mel_spectrogram[:, :, frame_offset:frame_offset + num_frames]
